from functools import partial

import autoarray as aa
import autofit as af

from autocti.charge_injection.imaging.imaging import ImagingCI


def _imaging_ci_list_from(fit: af.Fit, use_dataset_full: bool = False):
    """
    Returns a `ImagingCI` object from a `PyAutoFit` sqlite database `Fit` object.

    The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

    - The charge injection dataset data as a .fits file (`dataset/data.fits`).
    - The noise-map as a .fits file (`dataset/noise_map.fits`).
    - The pre CTI data as a .fits file (`dataset/pre_cti_data.fits`).
    - The cosmic ray map (if included) as a .fits file (`dataset/cosmic_ray_map.fits`).
    - The layout of the `ImagingCI` data structure used in the fit (`dataset/layout.json`).
    - The settings dictionary of the data (if used) as a .json file (`dataset/settings_dict.json`).
    - The mask used to mask the `ImagingCI` data structure in the fit (`dataset/mask.fits`).

    Each individual attribute can be loaded from the database via the `fit.value()` method.

    This method combines all of these attributes and returns a `ImagingCI` object, has the mask applied to the
    `ImagingCI` data structure and its settings updated to the values used by the model-fit.

    If multiple `ImagingCI` objects were fitted simultaneously via analysis summing, the `fit.child_values()` method
    is instead used to load lists of the data, noise-map, pre CTI data, layout and mask and combine them into a list of
    `ImagingCI` objects.

    If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
    to the database, the input `use_dataset_full` can be switched in to load instead the full `ImagingCI` objects.

    Parameters
    ----------
    fit
        A `PyAutoFit` `Fit` object which contains the results of a model-fit as an entry in a sqlite database.
    use_dataset_full
        If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
        to the database, the input `use_dataset_full` can be switched in to load instead the full `ImagingCI` objects.
    """

    if use_dataset_full:
        folder = "dataset_full"
    else:
        folder = "dataset"

    if not fit.children:
        fit_list = [fit]
    else:
        fit_list = fit.children

    dataset_list = []

    for fit in fit_list:
        layout = fit.value(name=f"{folder}.layout")

        data = aa.Array2D.from_primary_hdu(primary_hdu=fit.value(name=f"{folder}.data"))
        noise_map = aa.Array2D.from_primary_hdu(
            primary_hdu=fit.value(name=f"{folder}.noise_map")
        )
        pre_cti_data = aa.Array2D.from_primary_hdu(
            primary_hdu=fit.value(name=f"{folder}.pre_cti_data")
        )
        try:
            cosmic_ray_map = aa.Array2D.from_primary_hdu(
                primary_hdu=fit.value(name=f"{folder}.cosmic_ray_map")
            )
        except AttributeError:
            cosmic_ray_map = None

        settings_dict = fit.value(name="dataset.settings_dict")

        dataset = ImagingCI(
            data=data,
            noise_map=noise_map,
            pre_cti_data=pre_cti_data,
            cosmic_ray_map=cosmic_ray_map,
            settings_dict=settings_dict,
            layout=layout,
        )

        mask = aa.Mask2D.from_primary_hdu(primary_hdu=fit.value(name=f"{folder}.mask"))

        dataset_list.append(dataset.apply_mask(mask=mask))

    return dataset_list


class ImagingCIAgg:
    def __init__(self, aggregator: af.Aggregator, use_dataset_full: bool = False):
        """
        Interfaces with an `PyAutoFit` aggregator object to create instances of `ImagingCI` objects from the results
        of a model-fit.

        The results of a model-fit can be stored in a sqlite database, including the following attributes of the fit:

        - The charge injection dataset data as a .fits file (`dataset/data.fits`).
        - The noise-map as a .fits file (`dataset/noise_map.fits`).
        - The pre CTI data as a .fits file (`dataset/pre_cti_data.fits`).
        - The cosmic ray map (if included) as a .fits file (`dataset/cosmic_ray_map.fits`).
        - The layout of the `ImagingCI` data structure used in the fit (`dataset/layout.json`).
        - The settings dictionary of the data (if used) as a .json file (`dataset/settings_dict.json`).
        - The mask used to mask the `ImagingCI` data structure in the fit (`dataset/mask.fits`).

        The `aggregator` contains the path to each of these files, and they can be loaded individually. This class
        can load them all at once and create an `ImagingCI` object via the `_dataset_1d_from` method.

        This class's methods returns generators which create the instances of the `ImagingCI` objects. This ensures
        that large sets of results can be efficiently loaded from the hard-disk and do not require storing all
        `ImagingCI` instances in the memory at once.

        For example, if the `aggregator` contains 3 model-fits, this class can be used to create a generator which
        creates instances of the corresponding 3 `ImagingCI` objects.

        If multiple `ImagingCI` objects were fitted simultaneously via analysis summing, the `fit.child_values()`
        method is instead used to load lists of the data, noise-map, pre CTI data, layout and mask and combine them
        into a list of `ImagingCI` objects.

        If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
        to the database, the input `use_dataset_full` can be switched in to load instead the full `ImagingCI` objects.

        This can be done manually, but this object provides a more concise API.

        Parameters
        ----------
        aggregator
            A `PyAutoFit` aggregator object which can load the results of model-fits.
        use_dataset_full
            If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore
            accessible to the database, the input `use_dataset_full` can be switched in to load instead the
            full `ImagingCI` objects.
        """
        self.aggregator = aggregator
        self.use_dataset_full = use_dataset_full

    def dataset_list_gen_from(
        self,
    ):
        """
        Returns a generator of `ImagingCI` objects from an input aggregator.

        See `__init__` for a description of how the `ImagingCI` objects are created by this method.

        If a `dataset_full` is input into the `Analysis` class when a model-fit is performed and therefore accessible
        to the database, the input `use_dataset_full` can be switched in to load instead the full `ImagingCI` objects.
        """
        func = partial(_imaging_ci_list_from, use_dataset_full=self.use_dataset_full)
        return self.aggregator.map(func=func)



from autofit.non_linear import result


class Result(result.Result):
    def __init__(self, samples, analysis):
        """
        The result of a phase
        """
        super().__init__(samples=samples)

        self.analysis = analysis

    @property
    def clocker(self):
        return self.analysis.clocker


class ResultDataset(Result):
    @property
    def max_log_likelihood_fit(self):
        return self.analysis.fit_via_instance_from(instance=self.instance)

    @property
    def mask(self):
        return self.max_log_likelihood_fit.mask



import math
import numpy as np
from typing import List, Union

from arcticpy import CCDPhase
from arcticpy import TrapInstantCapture

from autocti.instruments.euclid import euclid_util

from autoarray.layout import layout_util
from autoarray.structures.arrays.uniform_2d import Array2D

from autocti.clocker.two_d import Clocker2D
from autocti.model.model_util import CTI2D

from autocti.charge_injection.layout import Layout2DCI
from autocti.charge_injection.imaging.simulator import SimulatorImagingCI

from autocti.charge_injection import ci_util

"""
Note on the rotations of arrays:

The function 'non_uniform_array_for_ou_sim' returns an array that is rotated according to the iquad parameter
(which ELVIS uses to define which quadrant the data corresponds to). This uniquely defines the orientation of the 
ndarray necessary to add parallel and serial CTI in the correct direction. These are defined accoridng to:

    http://euclid.esac.esa.int/dm/dpdd/latest/le1dpd/dpcards/le1_visrawframe.html

This function will return a array, which OU-Sim will then add the following effects too:

    - Add cosmic rays.
    - Bias.
    - Non-linearity.
    - Crosstalk?

The addition of CTI can then be performed using either the 'add_cti_to_array_for_ou_sim' function, or the standard
function OU-Sim use to add CTI (I guess they ultimately both flow through arctic in an identical way, albeit our
code uses SWIG and omits the need to define arCTIc parameter files.

Due to the rotations performed above, this means all images produced will be ndarrays oriented in the same way. I 
am not clear on how standard ELViS data products are oriented, but it may be that we require rotations before writing
them to fits to oriented them in the way they are observed (VIS_CTI has tools for this, as I'm sure ELViS does too).
"""


def quadrant_id_from(iquad: int) -> str:
    """
    The ELVIS simulator uses the `iquad` parameter to determine how images are rotated before clocking via arctic.

    This script converts this parameter to the `quadrant_id` used by PyAutoCTI, which in turn gives the appropriate
    `roe_corner` for rotation.

    The mapping of `iquad` to `quadrant_id` does not depend on the CCD id, because ELVIS has already performed
    extractions / rotations on the quadrant data beforehand.

    Parameters
    ----------
    iquad
        The ELVIS parameter defining the quadrant of the data and therefore the rotateion before arctic clocking.

    Returns
    -------
    str
        The quadrant ID string, which is either E, G, H or G
    """

    if iquad == 0:
        return "E"
    elif iquad == 1:
        return "F"
    elif iquad == 2:
        return "H"
    elif iquad == 3:
        return "G"


def injection_total_from(
    injection_start: int,
    injection_end: int,
    injection_on: int,
    injection_off: int,
):
    """
    The total number of charge injection regions for these electronics settings.
    """

    injection_range = injection_end - injection_start

    for injection_total in range(100):
        total_pixels = math.floor(
            (injection_total + 1) * (injection_on) + injection_total * injection_off
        )

        if total_pixels > injection_range:
            return injection_total


def charge_injection_array_from(
    ccd_id: str,
    quadrant_id: str,
    injection_norm: float,
    injection_start: int = 16,
    injection_end: int = 2066,
    injection_on: int = 200,
    injection_off: int = 200,
    parallel_size: int = 2086,
    serial_size: int = 2128,
    serial_prescan_size: int = 51,
    serial_overscan_size: int = 29,
    pixel_scales=0.1,
    use_non_uniform_pattern: bool = True,
    ci_seed: int = -1,
) -> Union[np.ndarray, Array2D]:
    """
    Returns a charge injection line image suitable for OU-SIM to run through the ElVIS simulator.

    By default, this array has dimensions (2086, 2128), representing a Euclid quadrant with a serial prescan of
    size 51 pixels, a serial overscan with 29 pixels and parallel overscan with 20 pixels.

    The charge injection line pattern is simulated using the VIS_CTI Processing element, and includes
    effects such as a non-uniform charge injection pattern. The charge injection is simulated in 3 distinct
    regions on the quadrant.

    This function assumes the same orientation for the charge injection line image, irrespective of the Euclid
    CCDPhase ID and Quadrant ID. The orientation that it assumes has arctic clock the ndarray towards [0, 0]. However,
    based on an input CCDPhase ID and Quadrant ID, the array is rotated to match Euclid clocking.

    Parameters
    ----------
    ccd_id
        The CCDPhase ID of Euclid (runs 1 through 6)
    quadrant_id
        The quadrant id (E, F, G, H)
    injection_norm
        The normalization of the charge injection region.
    parallel_size
        The size of the image in the parallel clocking direction (e.g. number of rows).
    serial_size
        The size of the image in the serial clocking direction (e.g. number of columns).
    serial_overscan_size
        The size of the serial overscan
    pixel_scales
        The arc-second to pixel scale conversion factor.

    Returns
    -------
    ndarray
        The charge injection line image oriented to match a given Euclid quadrant.
    """
    shape_native = (parallel_size, serial_size)

    injection_total = injection_total_from(
        injection_start=injection_start,
        injection_end=injection_end,
        injection_on=injection_on,
        injection_off=injection_off,
    )

    """
    Specify the charge injection regions on the CCDPhase, which in this case is 5 equally spaced rectangular blocks.
    
    At the end of this function the ndarray containing the charge injection data is rotated based on the quadrant_id.
    We therefore do not need rotated `regions_ci`'s from the function below, and input `roe_corner=(1,0)`, which 
    corresponds to quadrant E which is never rotated.
    """

    regions_ci = ci_util.region_list_ci_via_electronics_from(
        injection_start=injection_start,
        injection_on=injection_on,
        injection_off=injection_off,
        injection_total=injection_total,
        parallel_size=parallel_size,
        serial_size=serial_size,
        serial_prescan_size=serial_prescan_size,
        serial_overscan_size=serial_overscan_size,
        roe_corner=(1, 0),
    )

    """
    Use the charge injection normalization_list and regions to create `Layout2DCI` of every image we'll simulate.
    """
    layout = Layout2DCI(shape_2d=shape_native, region_list=regions_ci)

    """
    The simulator object creates simulations of charge injeciton imaging.
    """
    """
    Create every pre-cti charge injection image using each `Layout2DCI`
    """
    if use_non_uniform_pattern:
        simulator = SimulatorImagingCI(
            pixel_scales=pixel_scales,
            norm=injection_norm,
            row_slope=0.0,
            column_sigma=100.0,
            max_norm=200000,
            ci_seed=ci_seed,
        )

        pre_cti_data = simulator.pre_cti_data_non_uniform_from(layout=layout)
    else:
        simulator = SimulatorImagingCI(pixel_scales=pixel_scales, norm=injection_norm)

        pre_cti_data = simulator.pre_cti_data_uniform_from(layout=layout)

    """
    The OU-SIM parameter iquad defines the quadrant_id of the data (e.g. "E", "F", "G" or "H").
    """
    # quadrant_id = quadrant_id_from(iquad=iquad)

    roe_corner = euclid_util.roe_corner_from(ccd_id=ccd_id, quadrant_id=quadrant_id)

    """
    The array is rotated back to its original reference frame via the roe_corner, so other OU-Sim processing 
    works correctly.
    """
    return layout_util.rotate_array_via_roe_corner_from(
        array=pre_cti_data.native, roe_corner=roe_corner
    )


def add_cti_to_pre_cti_data(
    pre_cti_data: Union[np.ndarray, Array2D],
    ccd_id: str,
    quadrant_id: str,
    clocker: Clocker2D,
    parallel_trap_list: List[TrapInstantCapture],
    parallel_ccd: CCDPhase,
    serial_trap_list: List[TrapInstantCapture],
    serial_ccd: CCDPhase,
) -> Union[np.ndarray, Array2D]:
    #  quadrant_id = quadrant_id_from(iquad=iquad)

    roe_corner = euclid_util.roe_corner_from(ccd_id=ccd_id, quadrant_id=quadrant_id)

    pre_cti_data = layout_util.rotate_array_via_roe_corner_from(
        array=pre_cti_data, roe_corner=roe_corner
    )

    """
    The `Clocker` models the CCDPhase read-out, including CTI.

    For parallel clocking, we use 'charge injection mode' which transfers the charge of every pixel over the full CCDPhase.
    """

    """
    The CTI model used by arCTIc to add CTI to the input image in the parallel direction, which contains: 

        - 2 `TrapInstantCapture` species in the parallel direction.
        - A simple CCDPhase volume beta parametrization.
        - 3 `TrapInstantCapture` species in the serial direction.
        - A simple CCDPhase volume beta parametrization.
    """

    cti = CTI2D(
        parallel_trap_list=parallel_trap_list,
        parallel_ccd=parallel_ccd,
        serial_trap_list=serial_trap_list,
        serial_ccd=serial_ccd,
    )

    post_cti_data = clocker.add_cti(data=pre_cti_data, cti=cti)

    return layout_util.rotate_array_via_roe_corner_from(
        array=post_cti_data, roe_corner=roe_corner
    )



import copy
import numpy as np
from typing import List

import autoarray as aa

from autoarray.dataset.imaging.simulator import SimulatorImaging

from autocti.charge_injection.imaging.imaging import ImagingCI
from autocti.charge_injection.layout import Layout2DCI
from autocti.clocker.two_d import Clocker2D
from autocti.extract.settings import SettingsExtract
from autocti.model.model_util import CTI2D

from typing import Optional


class SimulatorImagingCI(SimulatorImaging):
    def __init__(
        self,
        pixel_scales: aa.type.PixelScales,
        norm: float,
        max_norm: float = np.inf,
        column_sigma: Optional[float] = None,
        row_slope: Optional[float] = 0.0,
        non_uniform_norm_limit=None,
        read_noise: Optional[float] = None,
        charge_noise: Optional[float] = None,
        noise_if_add_noise_false: float = 0.1,
        noise_seed: int = -1,
        ci_seed: int = -1,
    ):
        """A class representing a Imaging observation, using the shape of the image, the pixel scale,
        psf, exposure time, etc.

        Parameters
        ----------
        exposure_time_map
            The exposure time of an observation using this data_type.
        """

        super().__init__(
            read_noise=read_noise,
            exposure_time=1.0,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

        self.pixel_scales = pixel_scales

        self.norm = norm
        self.max_norm = max_norm
        self.column_sigma = column_sigma
        self.row_slope = row_slope
        self.non_uniform_norm_limit = non_uniform_norm_limit
        self.charge_noise = charge_noise

        self.ci_seed = ci_seed

    @property
    def _ci_seed(self) -> int:
        if self.ci_seed == -1:
            return np.random.randint(0, int(1e9))
        return self.ci_seed

    def median_list_from(self, total_columns: int) -> List[float]:
        np.random.seed(self._ci_seed)

        injection_norm_list = []

        for column_number in range(total_columns):
            injection_norm = 0

            while injection_norm <= 0 or injection_norm >= self.max_norm:
                injection_norm = np.random.normal(self.norm, self.column_sigma)

            injection_norm_list.append(injection_norm)

        return injection_norm_list

    def injection_norm_list_with_limit_from(self, total_columns: int) -> List[float]:
        injection_norm_list = self.median_list_from(
            total_columns=self.non_uniform_norm_limit
        )

        injection_norm_limited_list = []

        for i in range(total_columns):
            injection_norm = np.random.choice(injection_norm_list)

            injection_norm_limited_list.append(injection_norm)

        return injection_norm_limited_list

    def pre_cti_data_uniform_from(self, layout: Layout2DCI) -> aa.Array2D:
        """
        Use this charge injection layout_ci to generate a pre-cti charge injection image. This is performed by \
        going to its charge injection regions and adding the charge injection normalization value.

        Parameters
        ----------
        shape_native
            The image_shape of the pre_cti_datas to be created.
        """
        return layout.pre_cti_data_uniform_from(
            norm=self.norm, pixel_scales=self.pixel_scales
        )

    def pre_cti_data_non_uniform_from(self, layout: Layout2DCI) -> aa.Array2D:
        """
        Use this charge injection layout to generate a pre-cti charge injection image. This is performed by going
        to its charge injection regions and adding an input normalization value to each column, which are
        passed in as a list.

        For one column of a non-uniform charge injection pre-cti image, this function assumes that each non-uniform
        charge  injection region has the same overall normalization value. This assumes the spikes / troughs in the
        injection current that cause non-uniformity occur in an identical fashion for each charge injection region.

        Non-uniformity across the rows of a charge injection layout_ci is due to a drop-off in voltage in the current.
        Therefore, it appears smooth and be modeled as an analytic function, which this code assumes is a
        power-law with slope `row_slope`.

        Parameters
        ----------
        shape_native
            The image_shape of the pre_cti_datas to be created.
        row_slope
            The power-law slope of non-uniformity in the row charge injection profile.
        ci_seed
            Input ci_seed for the random number generator to give reproducible results. A new ci_seed is always used for each \
            pre_cti_datas, ensuring each non-uniform ci_region has the same column non-uniformity layout_ci.
        """

        for region in layout.region_list:
            if self.non_uniform_norm_limit is None:
                injection_norm_list = self.median_list_from(
                    total_columns=region.total_columns
                )
            else:
                injection_norm_list = self.injection_norm_list_with_limit_from(
                    total_columns=region.total_columns
                )

        return layout.pre_cti_data_non_uniform_from(
            injection_norm_list=injection_norm_list,
            pixel_scales=self.pixel_scales,
            row_slope=self.row_slope,
        )

    def via_layout_from(
        self,
        layout: Layout2DCI,
        clocker: Optional[Clocker2D],
        cti: Optional[CTI2D],
        cosmic_ray_map: Optional[aa.Array2D] = None,
    ) -> ImagingCI:
        """Simulate a charge injection image, including effects like noises.

        Parameters
        ----------
        pre_cti_data
        cosmic_ray_map
            The dimensions of the output simulated charge injection image.
        frame_geometry
            The quadrant geometry of the simulated image, defining where the parallel / serial overscans are and \
            therefore the direction of clocking and rotations before input into the cti algorithm.
        layout
            The charge injection layout_ci (regions, normalization, etc.) of the charge injection image.
        cti_params
            The CTI model parameters (trap density, trap release_timescales etc.).
        clocker
            The settings that control the cti clocking algorithm (e.g. ccd well_depth express option).
        read_noise
            The FWHM of the Gaussian read-noises added to the image.
        noise_seed
            Seed for the read-noises added to the image.
        """

        if self.column_sigma is not None:
            pre_cti_data = self.pre_cti_data_non_uniform_from(layout=layout)
        else:
            pre_cti_data = self.pre_cti_data_uniform_from(layout=layout)

        return self.via_pre_cti_data_from(
            pre_cti_data=pre_cti_data.native,
            layout=layout,
            clocker=clocker,
            cti=cti,
            cosmic_ray_map=cosmic_ray_map,
        )

    def via_pre_cti_data_from(
        self,
        pre_cti_data: aa.Array2D,
        layout: Layout2DCI,
        clocker: Optional[Clocker2D],
        cti: Optional[CTI2D],
        cosmic_ray_map: Optional[aa.Array2D] = None,
    ) -> ImagingCI:
        pre_cti_data = pre_cti_data.native

        if cosmic_ray_map is not None:
            pre_cti_data += cosmic_ray_map.native

        if self.charge_noise is not None:
            pre_cti_data = layout.extract.parallel_fpr.add_gaussian_noise_to(
                array=pre_cti_data,
                noise_sigma=self.charge_noise,
                noise_seed=self.noise_seed,
                settings=SettingsExtract(
                    pixels_from_end=layout.extract.parallel_fpr.total_rows_min
                ),
            )

        if cti is not None:
            post_cti_data = clocker.add_cti(data=pre_cti_data, cti=cti)
        else:
            post_cti_data = copy.copy(pre_cti_data)

        if cosmic_ray_map is not None:
            pre_cti_data -= cosmic_ray_map.native

        return self.via_post_cti_data_from(
            post_cti_data=post_cti_data,
            pre_cti_data=pre_cti_data,
            layout=layout,
            cosmic_ray_map=cosmic_ray_map,
        )

    def via_post_cti_data_from(
        self,
        post_cti_data: aa.Array2D,
        pre_cti_data: aa.Array2D,
        layout: Layout2DCI,
        cosmic_ray_map: Optional[aa.Array2D] = None,
    ) -> ImagingCI:
        if self.read_noise is not None:
            ci_image = aa.preprocess.data_with_gaussian_noise_added(
                data=post_cti_data, sigma=self.read_noise, seed=self.noise_seed
            )

            ci_image = aa.Array2D.no_mask(
                values=ci_image, pixel_scales=self.pixel_scales
            )

            ci_noise_map = (
                self.read_noise
                * aa.Array2D.ones(
                    shape_native=layout.shape_2d, pixel_scales=self.pixel_scales
                ).native
            )
        else:
            ci_image = post_cti_data
            ci_noise_map = aa.Array2D.full(
                fill_value=self.noise_if_add_noise_false,
                shape_native=layout.shape_2d,
                pixel_scales=self.pixel_scales,
            ).native

        return ImagingCI(
            data=aa.Array2D.no_mask(
                values=ci_image.native, pixel_scales=self.pixel_scales
            ),
            noise_map=aa.Array2D.no_mask(
                values=ci_noise_map, pixel_scales=self.pixel_scales
            ),
            pre_cti_data=aa.Array2D.no_mask(
                values=pre_cti_data.native, pixel_scales=self.pixel_scales
            ),
            cosmic_ray_map=cosmic_ray_map,
            layout=layout,
        )



from typing import List

import autoarray as aa

from autocti.extract.two_d.parallel.abstract import Extract2DParallel
from autocti.extract.settings import SettingsExtract

from autocti.extract.two_d import extract_2d_util


class Extract2DParallelPreInjection(Extract2DParallel):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region2D]:
        """
        Returns a list of the 2D parallel regions before the first injected signal (e.g. before the FPR) from
        the `region_list` containing signal  (e.g. the charge injection regions of charge injection data), between
        two input `pixels` indexes.

        Negative pixel values can be input into the `pixels` tuple, whereby columns in front of the region in front of
        the parallel FPRs are extracted.

        A 2D region is defined following the convention:

        (top-row, bottom-row, left-column, right-column) = (y0, y1, x0, x1)

        The pre injection region is defined to contain all pixels before the FPR, not including the rows of pixels
        in the serial prescan and overscans. This is because the region is used to estimate and subtract constant
        background signals across the full CCD (e.g. bias, stray light, etc.). The x0 and x1 coordinates are therefore
        updated to include the serial prescan and overscan regions.

        The y0 and y1 coordinates are updated based on the `pixels` input, but also computed based on where the
        injection occurs (and therefore are before the FPRs).

        The diagram below illustrates the extraction for `pixels=(0, 1)`:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = parallel / serial charge injection region trail
        [ppippippipp] The pre injection region to be extracted

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
           [...][ttttttttttttttttttttt][sss]
           [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        |  [...][1c1c1cc1c1cc1ccc1cc1][sss]    |
        |  [...][ttttttttttttttttttttt][sss]    | Direction
        Par [...][ttttttttttttttttttttt][sss]    | of
        |  [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
        |/  [...][cc0ccc0cccc0cccc0cccc][sss]    \/
        |/  [...][ppippippippipppipppip][sss]

        []     [=====================]
               <---------Ser--------

        The extracted regions correspond to the single region before the first parallel FPR:

        region_list[0] = [0, 4, 0, 1] (serial prescan is 3 pixels)
        region_list[1] = [0, 4, 0, 1] (serial prescan is 3 pixels)

        For `pixels=(0,1)` the extracted arrays returned via the `array_2d_list_from()` function keep the first
        parallel FPR of each charge injection region:

        array_2d_list[0] = [ppippippippipppipppip]

        Parameters
        ----------
        settings
           The settings used to extract the parallel FPRs, which for example include the `pixels` tuple specifying the
           range of pixel rows they are extracted between.
        """

        pixels = settings.pixels
        pixels_from_end = settings.pixels_from_end

        if pixels_from_end is not None:
            y0_min = min([region.y0 for region in self.region_list])

            pixels = (y0_min - pixels_from_end, y0_min)

        y0 = pixels[1]

        return [
            aa.Region2D(
                region=(pixels[0], y0, self.region_list[0].x0, self.region_list[0].x1)
            )
        ]



import copy
from typing import Dict, Optional


import autoarray as aa

from autocti.preloads import Preloads
from autocti.charge_injection.imaging.imaging import ImagingCI


class FitImagingCI(aa.FitImaging):
    def __init__(
        self,
        dataset: ImagingCI,
        post_cti_data: aa.Array2D,
        hyper_noise_scalar_dict: Optional[Dict] = None,
        preloads: Preloads = Preloads(),
    ):
        """
        Fit a charge injection ci_data-set with a model cti image, also scalng the noises within a Bayesian
        framework.

        Parameters
        ----------
        dataset
            The charge injection image that is fitted.
        post_cti_data
            The `pre_cti_data` with cti added to it via the clocker and a CTI model.
        hyper_noise_scalar_dict
            The hyper_ci-parameter(s) which the noise_scaling_map_dict_list is multiplied by to scale the noise-map.
        """

        super().__init__(dataset=dataset, use_mask_in_fit=True)

        self.post_cti_data = post_cti_data
        self.hyper_noise_scalar_dict = hyper_noise_scalar_dict
        self.noise_scaling_map_dict = dataset.noise_scaling_map_dict
        self.preloads = preloads

    @property
    def imaging_ci(self) -> ImagingCI:
        return self.dataset

    @property
    def noise_map(self) -> aa.Array2D:
        if self.hyper_noise_scalar_dict is not None:
            return hyper_noise_map_from(
                hyper_noise_scalar_dict=self.hyper_noise_scalar_dict,
                noise_scaling_map_dict=self.noise_scaling_map_dict,
                noise_map=self.dataset.noise_map,
            )

        return self.dataset.noise_map

    @property
    def model_data(self) -> aa.Array2D:
        return self.post_cti_data

    @property
    def layout(self):
        return self.dataset.layout

    @property
    def pre_cti_data(self):
        return self.dataset.pre_cti_data

    @property
    def chi_squared(self) -> float:
        """
        Returns the chi-squared terms of the model data's fit to an dataset, by summing the chi-squared-map.

        If the dataset includes a noise covariance matrix, this is used instead to account for covariance in the
        goodness-of-fit.

        The standard chi-squared calculation in PyAutoArray computes the `chi-squared` from the `residual_map`
        and `chi_squared_map`, which requires that the `ndarrays` which are used to do this are created and stored
        in memory. For charge injection imaging, the large datasets mean this can be computationally slow.

        This function computes the `chi_squared` directly from the data, avoiding the need to store the data in memory
        and offering faster tune times.
        """

        return aa.util.fit.chi_squared_with_mask_fast_from(
            data=self.dataset.data,
            noise_map=self.noise_map,
            mask=self.mask,
            model_data=self.model_data,
        )

    @property
    def noise_normalization(self) -> float:
        """
        Returns the noise-map normalization term of the noise-map, summing the noise_map value in every pixel as:

        [Noise_Term] = sum(log(2*pi*[Noise]**2.0))
        """

        if self.preloads.noise_normalization is not None:
            return self.preloads.noise_normalization

        return aa.util.fit.noise_normalization_with_mask_from(
            noise_map=self.noise_map, mask=self.mask
        )

    @property
    def chi_squared_map_of_regions_ci(self):
        return self.layout.extract.regions_array_2d_from(array=self.chi_squared_map)

    @property
    def chi_squared_map_of_parallel_eper(self):
        return self.layout.extract.parallel_eper.array_2d_from(
            array=self.chi_squared_map
        )

    @property
    def chi_squared_map_of_serial_eper(self):
        return self.layout.extract.serial_eper.array_2d_from(array=self.chi_squared_map)

    @property
    def chi_squared_map_of_serial_overscan_no_eper(self):
        return self.layout.extract.serial_overscan_above_eper_array_2d_from(
            array=self.chi_squared_map
        )


def hyper_noise_map_from(hyper_noise_scalar_dict, noise_scaling_map_dict, noise_map):
    """
    For a noise-map, use the model hyper noise and noise-scaling maps to compute a scaled noise-map.

    Parameters
    ----------
    hyper_noise_scalar_dict
    noise_scaling_map_dict
    noise_map : imaging.NoiseMap or ndarray
        An arrays describing the RMS standard deviation error in each pixel, preferably in units of electrons per
        second.
    """

    noise_map = copy.copy(noise_map)

    for key, hyper_noise_scalar in hyper_noise_scalar_dict.items():
        noise_map += hyper_noise_scalar.scaled_noise_map_from(
            noise_scaling=noise_scaling_map_dict[key]
        )

    return noise_map



from copy import deepcopy
import numpy as np
from typing import List, Optional, Tuple

import autoarray as aa

from autocti.extract.settings import SettingsExtract

from autocti import exc


class Layout2D(aa.Layout2D):
    def __init__(
        self,
        shape_2d: Tuple[int, int],
        region_list: aa.type.Region2DList,
        original_roe_corner: Tuple[int, int] = (1, 0),
        parallel_overscan: Optional[aa.type.Region2DLike] = None,
        serial_prescan: Optional[aa.type.Region2DLike] = None,
        serial_overscan: Optional[aa.type.Region2DLike] = None,
    ):
        """
        A charge injection layout, which defines the regions charge injections appear on a charge injection image.

        It also contains over regions of the image, for example the serial prescan, overscan and paralle overscan.

        Parameters
        ----------
        shape_2d
            The two dimensional shape of the charge injection imaging, corresponding to the number of rows (pixels
            in parallel direction) and columns (pixels in serial direction).
        region_list
            Integer pixel coordinates specifying the corners of each charge injection region (top-row, bottom-row,
            left-column, right-column).
        original_roe_corner
            The original read-out electronics corner of the charge injeciton imaging, which is internally rotated to a
            common orientation in **PyAutoCTI**.
        parallel_overscan
            Integer pixel coordinates specifying the corners of the parallel overscan (top-row, bottom-row,
            left-column, right-column).
        serial_prescan
            Integer pixel coordinates specifying the corners of the serial prescan (top-row, bottom-row,
            left-column, right-column).
        serial_overscan
            Integer pixel coordinates specifying the corners of the serial overscan (top-row, bottom-row,
            left-column, right-column).
        electronics
            The charge injection electronics parameters of the image (e.g. the IG1 and IG2 voltages).
        """

        self.region_list = list(map(aa.Region2D, region_list))

        for region in self.region_list:
            if region.y1 > shape_2d[0] or region.x1 > shape_2d[1]:
                raise exc.LayoutException(
                    "The charge injection layout_ci regions are bigger than the image image_shape"
                )

        super().__init__(
            shape_2d=shape_2d,
            original_roe_corner=original_roe_corner,
            parallel_overscan=parallel_overscan,
            serial_prescan=serial_prescan,
            serial_overscan=serial_overscan,
        )

    @property
    def extract(self):
        from autocti.extract.two_d.master import Extract2DMaster

        return Extract2DMaster(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self.parallel_overscan,
            serial_prescan=self.serial_prescan,
            serial_overscan=self.serial_overscan,
        )

    def layout_extracted_from(
        self, extraction_region: aa.type.Region2DLike
    ) -> "Layout2D":
        """
        The charge injection layout after an extraction is performed on its associated charge injection image, where
        the extraction is defined by a region of pixel coordinates:

        (top-row, bottom-row, left-column, right-column) = (y0, y1, x0, x1)

        For example, if a charge injection region occupies the pixels (0, 10, 0, 5) on a 100 x 100 charge injection
        image, and the first 5 columns of this charge injection image are extracted to create a 100 x 5 image, the
        new charge injection region will be (0, 20, 0, 5).

        Parameters
        ----------
        extraction_region
            The (y0, y1, x0, x1) pixel coordinates defining the region which the layout is extracted from.
        """

        layout = super().layout_extracted_from(extraction_region=extraction_region)

        region_list = [
            aa.util.layout.region_after_extraction(
                original_region=region, extraction_region=extraction_region
            )
            for region in self.region_list
        ]

        return self.__class__(
            original_roe_corner=self.original_roe_corner,
            shape_2d=self.shape_2d,
            region_list=region_list,
            parallel_overscan=layout.parallel_overscan,
            serial_prescan=layout.serial_prescan,
            serial_overscan=layout.serial_overscan,
        )

    @property
    def parallel_rows_between_regions(self) -> List[int]:
        """
        Returns a list where each entry is the number of pixels a charge injection region and its neighboring
        charge injection region.
        """
        return self.extract.parallel_fpr.parallel_rows_between_regions

    @property
    def parallel_rows_within_regions(self) -> List[int]:
        """
        Returns a list where each entry is the number of pixels a charge injection region and its neighboring
        charge injection region.
        """
        return [region.y1 - region.y0 for region in self.region_list]

    @property
    def parallel_rows_to_array_edge(self) -> int:
        """
        The number of pixels from the edge of the parallel EPERs to the edge of the array.

        This is the number of pixels from the last charge injection FPR edge to the read out register and electronics
        and will include the parallel overscan if the CCD has one.
        """
        return self.extract.parallel_fpr.parallel_rows_to_array_edge

    @property
    def smallest_parallel_rows_between_ci_regions(self) -> int:
        """
        The smallest number of pixels between any two charge injection regions, or the distance of the last
        charge injection region to the edge of the charge injeciton image (e.g. in the direction away from the
        readout register and electronics).
        """
        parallel_rows_between_regions = self.parallel_rows_between_regions
        parallel_rows_between_regions.append(self.parallel_rows_to_array_edge)
        return np.min(parallel_rows_between_regions)

    @property
    def smallest_parallel_rows_within_ci_regions(self) -> int:
        """
        The smallest number of pixels within any of the charge injection regions.

        This can be used to extract the FPR of each charge injection region using a
        number of pixels which does not exceed the size of any.
        """
        return np.min(self.parallel_rows_within_regions)

    def with_extracted_regions(
        self, extraction_region: aa.type.Region2DLike
    ) -> "Layout2D":
        layout = deepcopy(self)

        extracted_region_list = list(
            map(
                lambda region: aa.util.layout.region_after_extraction(
                    original_region=region, extraction_region=extraction_region
                ),
                self.region_list,
            )
        )
        extracted_region_list = list(filter(None, extracted_region_list))
        if not extracted_region_list:
            extracted_region_list = None

        layout.region_list = extracted_region_list
        return layout

    def extract_region_from(self, array: aa.Array2D, region: str) -> aa.Array1D:
        if region == "parallel_fpr":
            return self.extract.parallel_fpr.binned_array_1d_from(
                array=array,
                settings=SettingsExtract(
                    pixels=(0, self.extract.parallel_fpr.total_rows_min)
                ),
            )
        elif region == "parallel_eper":
            return self.extract.parallel_eper.binned_array_1d_from(
                array=array,
                settings=SettingsExtract(
                    pixels=(0, self.smallest_parallel_rows_between_ci_regions)
                ),
            )
        elif region == "serial_fpr" or region == "fpr_non_uniformity":
            return self.extract.serial_fpr.binned_array_1d_from(
                array=array,
                settings=SettingsExtract(
                    pixels=(0, self.extract.serial_fpr.total_columns_min)
                ),
            )
        elif region == "serial_eper":
            return self.extract.serial_eper.binned_array_1d_from(
                array=array,
                settings=SettingsExtract(pixels=(0, self.serial_eper_pixels)),
            )
        else:
            raise exc.PlottingException(
                "The line region specified for the plotting of a line was invalid"
            )

    def extract_region_noise_map_from(
        self, array: aa.Array2D, region: str
    ) -> aa.Array1D:
        if region == "parallel_fpr":
            binned_noise_map_1d = self.extract.parallel_fpr.binned_array_1d_from(
                array=array,
                settings=SettingsExtract(
                    pixels=(0, self.extract.parallel_fpr.total_rows_min)
                ),
            )
            binned_noise_map_1d_total_pixels = (
                self.extract.parallel_fpr.binned_array_1d_total_pixels_from(
                    array=array,
                    settings=SettingsExtract(
                        pixels=(0, self.extract.parallel_fpr.total_rows_min)
                    ),
                )
            )
        elif region == "parallel_eper":
            binned_noise_map_1d = self.extract.parallel_eper.binned_array_1d_from(
                array=array,
                settings=SettingsExtract(
                    pixels=(0, self.smallest_parallel_rows_between_ci_regions)
                ),
            )
            binned_noise_map_1d_total_pixels = (
                self.extract.parallel_eper.binned_array_1d_total_pixels_from(
                    array=array,
                    settings=SettingsExtract(
                        pixels=(0, self.smallest_parallel_rows_between_ci_regions),
                    ),
                )
            )
        elif region == "serial_fpr" or region == "fpr_non_uniformity":
            binned_noise_map_1d = self.extract.serial_fpr.binned_array_1d_from(
                array=array,
                settings=SettingsExtract(
                    pixels=(0, self.extract.serial_fpr.total_columns_min)
                ),
            )
            binned_noise_map_1d_total_pixels = (
                self.extract.serial_fpr.binned_array_1d_total_pixels_from(
                    array=array,
                    settings=SettingsExtract(
                        pixels=(0, self.extract.serial_fpr.total_columns_min)
                    ),
                )
            )
        elif region == "serial_eper":
            binned_noise_map_1d = self.extract.serial_eper.binned_array_1d_from(
                array=array,
                settings=SettingsExtract(pixels=(0, self.serial_eper_pixels)),
            )
            binned_noise_map_1d_total_pixels = (
                self.extract.serial_eper.binned_array_1d_total_pixels_from(
                    array=array,
                    settings=SettingsExtract(pixels=(0, self.serial_eper_pixels)),
                )
            )
        else:
            raise exc.PlottingException(
                "The line region specified for the plotting of a line was invalid"
            )

        binned_noise_map_1d /= np.sqrt(binned_noise_map_1d_total_pixels)

        return binned_noise_map_1d



from typing import List, Optional, Tuple

from arcticpy import TrapInstantCapture

from autoconf.dictable import from_json, output_to_json

from autocti import exc


class AbstractSettingsCTI:
    def check_total_density_within_range_of_traps(
        self, total_density_range: Tuple[float, float], traps: List[TrapInstantCapture]
    ):
        if total_density_range is not None:
            total_density = sum([trap.density for trap in traps])

            if (
                total_density < total_density_range[0]
                or total_density > total_density_range[1]
            ):
                raise exc.PriorException

    @classmethod
    def from_json(cls, file_path):
        return from_json(file_path=file_path)

    def output_to_json(self, file_path):
        output_to_json(obj=self, file_path=file_path)


class SettingsCTI1D(AbstractSettingsCTI):
    def __init__(
        self,
        total_density_range: Optional[Tuple[float, float]] = None,
    ):
        """
        Controls the modeling settings of CTI clocking in 1D.

        For a CTI model-fit in 1D, the total density of all traps can be constrained to be within an input range of
        values.

        For example, if `total_density_range=(0.0, 0.5)`, the total density of all traps will be between 0.0 and 0.5
        electrons per pixel. This acts on the total density of traps, and therefore sums up the individual density
        of each trap species.

        Parameters
        ----------
        total_density_range
            The range of total trap density values allowed by the model-fit.
        """
        self.total_density_range = total_density_range

    def check_total_density_within_range(self, traps: List[TrapInstantCapture]):
        """
        Checks that the total density of traps is within the input range.

        This function receives as input lists of trap species and checks that their total density is within the
        range specified by the `total_density_range` attribute.

        If they are not in this range, an exception is raised, meaning that the model is resampled and this model is
        rejected.

        Parameters
        ----------
        traps
            The list of trap species whose densities are checked.
        """
        self.check_total_density_within_range_of_traps(
            total_density_range=self.total_density_range, traps=traps
        )


class SettingsCTI2D(AbstractSettingsCTI):
    def __init__(
        self,
        parallel_total_density_range: Optional[Tuple[float, float]] = None,
        serial_total_density_range: Optional[Tuple[float, float]] = None,
    ):
        """
        Controls the modeling settings of CTI clocking in 2D.

        For a CTI model-fit in 2D, the total density of all parallel and / or serial traps can be constrained to be
        within an input range of  values.

        For example, if `parallel_total_density_range=(0.0, 0.5)`, the total density of all parallel traps will be
        between 0.0 and 0.5  electrons per pixel. This acts on the total density of traps, and therefore sums up
        the individual density of each parallel trap species.

        Parameters
        ----------
        parallel_total_density_range
            The range of parallel total trap density values allowed by the model-fit.
        serial_total_density_range
            The range of serial total trap density values allowed by the model-fit.
        """
        self.parallel_total_density_range = parallel_total_density_range
        self.serial_total_density_range = serial_total_density_range

    def check_total_density_within_range(self, parallel_traps, serial_traps):
        """
        Checks that the total density of traps is within the input range.

        This function receives as input lists of parallel and / or serial trap species and checks that their total
        density is within the range specified by the `parallel_total_density_range` and `serial_total_density_range`
        attributes.

        If they are not in this range, an exception is raised, meaning that the model is resampled and this model is
        rejected.

        Parameters
        ----------
        parallel_traps
            The list of parallel trap species whose densities are checked.
        serial_traps
            The list of serial trap species whose densities are checked.
        """
        self.check_total_density_within_range_of_traps(
            total_density_range=self.parallel_total_density_range, traps=parallel_traps
        )

        self.check_total_density_within_range_of_traps(
            total_density_range=self.serial_total_density_range, traps=serial_traps
        )



def roe_corner_from(ccd_id, quadrant_id):
    row_index = ccd_id[-1]

    if (row_index in "123") and (quadrant_id == "E"):
        return (1, 0)
    elif (row_index in "123") and (quadrant_id == "F"):
        return (1, 1)
    elif (row_index in "123") and (quadrant_id == "G"):
        return (0, 1)
    elif (row_index in "123") and (quadrant_id == "H"):
        return (0, 0)
    elif (row_index in "456") and (quadrant_id == "E"):
        return (0, 1)
    elif (row_index in "456") and (quadrant_id == "F"):
        return (0, 0)
    elif (row_index in "456") and (quadrant_id == "G"):
        return (1, 0)
    elif (row_index in "456") and (quadrant_id == "H"):
        return (1, 1)



from typing import List, Optional, Tuple

import autoarray as aa

from autocti.extract.two_d.serial.abstract import Extract2DSerial
from autocti.extract.settings import SettingsExtract


class Extract2DSerialPrescan(Extract2DSerial):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region2D]:
        """
        Returns a list of the 2D serial prescan region, which is the parallel overscan input to the
        object, between two input `pixels` indexes (this is somewhat redundant information, but mimicks
        the `Extract` object API across all other `Extract` objects).

        (top-row, bottom-row, left-column, right-column) = (y0, y1, x0, x1)

        The serial prescan spans all columns of the image, thus the coordinates x0 and x1 do not change. y0 and y1
        are updated based on the `pixels` input.

        Negative pixel values can be input into the `pixels` tuple, whereby columns in front of the serial prescan are
        also extracted.

        The diagram below illustrates the extraction for `pixels=(0, 1)`:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = parallel overscan
        [ssssssssss] = serial overscan
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = serial / serial charge injection region trail

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
           [...][ttttttttttttttttttttt][sss]
           [...][f1f1ff1f1ff1ff1fff1ff][sss]
        |  [...][1f1f1ff1f1ff1fff1ff1][sss]    |
        |  [...][ttttttttttttttttttttt][sss]    | Direction
        Par [...][ttttttttttttttttttttt][sss]    | of
        |  [...][0fff0ffff0ffff0ffff0f][sss]    | clocking
        |/  [...][ff0fff0ffff0ffff0ffff][sss]    \/

        []     [=====================]
               <---------Ser--------

        The extracted regions correspond to the serial prescan [...] regions.

        For `pixels=(0,1)` the extracted arrays returned via the `array_2d_list_from()` function keep the first
        serial pixels across the columns of the prescan:

        array_2d_list[0] = [s]

        Parameters
        ----------
        settings
           The settings used to extract the serial region overscan from, which for example include the `pixels`
           tuple specifying the range of pixel columns they are extracted between.
        """

        pixels = settings.pixels

        if settings.pixels_from_end is not None:
            pixels = (
                self.serial_prescan.total_columns - settings.pixels_from_end,
                self.serial_prescan.total_columns,
            )

        serial_prescan_extract = aa.Region2D(
            (
                self.serial_prescan.y0,
                self.serial_prescan.y1,
                self.serial_prescan.x0 + pixels[0],
                self.serial_prescan.x0 + pixels[1],
            )
        )

        return [serial_prescan_extract]



import logging
import os
from os import path

from autoarray import exc
from autoarray.structures.arrays import array_2d_util

from autocti.instruments.acs.array_2d import Array2DACS
from autocti.instruments.acs.header import HeaderACS

from autocti.instruments.acs import acs_util

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel("INFO")


class ImageACS(Array2DACS):
    """
    The layout of an ACS array and image is given in `FrameACS`.

    This class handles specifically the image of an ACS observation, assuming that it contains specific
    header info.
    """

    @classmethod
    def from_fits(
        cls,
        file_path,
        quadrant_letter,
        bias_subtract_via_bias_file=False,
        bias_subtract_via_prescan=False,
        bias_file_path=None,
        use_calibrated_gain=True,
    ):
        """
        Use the input .fits file and quadrant letter to extract the quadrant from the full CCD, perform
        the rotations required to give correct arctic clocking and convert the image from units of COUNTS / CPS to
        ELECTRONS.

        See the docstring of the `FrameACS` class for a complete description of the HST FPA, quadrants and
        rotations.

        Also see https://github.com/spacetelescope/hstcal/blob/main/pkg/acs/calacs/acscte/dopcte-gen2.c#L418

        Parameters
        ----------
        file_path
            The full path of the file that the image is loaded from, including the file name and ``.fits`` extension.
        quadrant_letter
            The letter of the ACS quadrant the image is extracted from and loaded.
        bias_subtract_via_bias_file
            If True, the corresponding bias file of the image is loaded (via the name of the file in the fits header).
        bias_subtract_via_prescan
            If True, the prescan on the image is used to estimate a component of bias that is subtracted from the image.
        bias_file_path
            If `bias_subtract_via_bias_file=True`, this overwrites the path to the bias file instead of the default
            behaviour of using the .fits header.
        use_calibrated_gain
            If True, the calibrated gain values are used to convert from COUNTS to ELECTRONS.
        """

        hdu = acs_util.fits_hdu_via_quadrant_letter_from(
            quadrant_letter=quadrant_letter
        )

        header_sci_obj = array_2d_util.header_obj_from(file_path=file_path, hdu=0)
        header_hdu_obj = array_2d_util.header_obj_from(file_path=file_path, hdu=hdu)

        header = HeaderACS(
            header_sci_obj=header_sci_obj,
            header_hdu_obj=header_hdu_obj,
            hdu=hdu,
            quadrant_letter=quadrant_letter,
        )

        if header.header_sci_obj["TELESCOP"] != "HST":
            raise exc.ArrayException(
                f"The file {file_path} does not point to a valid HST ACS dataset."
            )

        if header.header_sci_obj["INSTRUME"] != "ACS":
            raise exc.ArrayException(
                f"The file {file_path} does not point to a valid HST ACS dataset."
            )

        array = array_2d_util.numpy_array_2d_via_fits_from(
            file_path=file_path, hdu=hdu, do_not_scale_image_data=True
        )

        array = header.array_original_to_electrons(
            array=array, use_calibrated_gain=use_calibrated_gain
        )

        if bias_subtract_via_bias_file:
            if bias_file_path is None:
                file_dir = os.path.split(file_path)[0]
                bias_file_path = path.join(file_dir, header.bias_file)

            bias = array_2d_util.numpy_array_2d_via_fits_from(
                file_path=bias_file_path, hdu=hdu, do_not_scale_image_data=True
            )

            header_sci_obj = array_2d_util.header_obj_from(
                file_path=bias_file_path, hdu=0
            )
            header_hdu_obj = array_2d_util.header_obj_from(
                file_path=bias_file_path, hdu=hdu
            )

            bias_header = HeaderACS(
                header_sci_obj=header_sci_obj,
                header_hdu_obj=header_hdu_obj,
                hdu=hdu,
                quadrant_letter=quadrant_letter,
            )

            if bias_header.original_units != "COUNTS":
                raise exc.ArrayException("Cannot use bias frame not in counts.")

            bias = bias * bias_header.calibrated_gain

        else:
            bias = None

        return cls.from_ccd(
            array_electrons=array,
            quadrant_letter=quadrant_letter,
            header=header,
            bias_subtract_via_prescan=bias_subtract_via_prescan,
            bias=bias,
        )



import numpy as np
from typing import Optional, Tuple

import autoarray as aa

from autocti.extract.settings import SettingsExtract
from autocti import exc


class Layout1D(aa.Layout1D):
    def __init__(
        self,
        shape_1d: Tuple[int],
        region_list: aa.type.Region1DList,
        prescan: aa.type.Region1DLike = None,
        overscan: aa.type.Region1DLike = None,
    ):
        """
        Abstract base class for a charge injection layout_ci, which defines the regions charge injections appears \
         on a charge-injection array, the input normalization and other properties.

        Parameters
        ----------
        region_list
            A list of the integer coordinates specifying the corners of each charge injection region \
            (top-row, bottom-row, left-column, right-column).
        """

        self.region_list = list(map(aa.Region1D, region_list))

        for region in self.region_list:
            if region.x1 > shape_1d[0]:
                raise exc.LayoutException(
                    "The charge injection layout_ci regions are bigger than the image image_shape"
                )

        from autocti.extract.one_d.master import Extract1DMaster

        self.extract = Extract1DMaster(
            shape_1d=shape_1d,
            region_list=region_list,
            prescan=prescan,
            overscan=overscan,
        )

        super().__init__(shape_1d=shape_1d, prescan=prescan, overscan=overscan)

    @property
    def parallel_rows_between_regions(self):
        return self.extract.fpr.parallel_rows_between_regions

    @property
    def trail_size_to_array_edge(self):
        return self.extract.fpr.trail_size_to_array_edge

    @property
    def smallest_eper_pixels_to_array_edge(self):
        parallel_rows_between_regions = self.parallel_rows_between_regions
        parallel_rows_between_regions.append(self.trail_size_to_array_edge)
        return np.min(parallel_rows_between_regions)

    def extract_region_from(self, array: aa.Array1D, region: Optional):
        if region is None:
            return array

        if region == "fpr":
            return self.extract.fpr.stacked_array_1d_from(
                array=array,
                settings=SettingsExtract(pixels=(0, self.extract.fpr.total_pixels_min)),
            )
        elif region == "eper":
            return self.extract.eper.stacked_array_1d_from(
                array=array,
                settings=SettingsExtract(
                    pixels=(0, self.smallest_eper_pixels_to_array_edge)
                ),
            )
        else:
            raise exc.PlottingException(
                "The line region specified for the plotting of a line was invalid"
            )



from typing import Dict, Tuple, Optional

from autoarray.structures.header import Header


class HeaderEuclid(Header):
    def __init__(
        self,
        header_sci_obj: Dict = None,
        header_hdu_obj: Dict = None,
        original_roe_corner: Tuple[int, int] = None,
        readout_offsets: Optional[Tuple] = None,
        ccd_id: Optional[str] = None,
        quadrant_id: Optional[str] = None,
    ):
        super().__init__(
            header_sci_obj=header_sci_obj,
            header_hdu_obj=header_hdu_obj,
            original_roe_corner=original_roe_corner,
            readout_offsets=readout_offsets,
        )

        self.ccd_id = ccd_id
        self.quadrant_id = quadrant_id

    @property
    def row_index(self) -> str:
        if self.ccd_id is not None:
            return self.ccd_id[-1]



import logging
from typing import List, Optional

from autoconf import conf
from autoconf.dictable import to_dict

import autoarray as aa
import autofit as af
from autoconf.dictable import output_to_json

from autocti.charge_injection.imaging.imaging import ImagingCI
from autocti.charge_injection.fit import FitImagingCI
from autocti.charge_injection.model.visualizer import VisualizerImagingCI
from autocti.charge_injection.model.result import ResultImagingCI
from autocti.clocker.two_d import Clocker2D
from autocti.charge_injection.hyper import HyperCINoiseCollection
from autocti.model.settings import SettingsCTI2D
from autocti.preloads import Preloads

from autocti import exc

logger = logging.getLogger(__name__)

logger.setLevel(level="INFO")


class AnalysisImagingCI(af.Analysis):
    def __init__(
        self,
        dataset: ImagingCI,
        clocker: Clocker2D,
        settings_cti: SettingsCTI2D = SettingsCTI2D(),
        dataset_full: Optional[ImagingCI] = None,
    ):
        """
        Fits a CTI model to a charge injection imaging dataset via a non-linear search.

        The `Analysis` class defines the `log_likelihood_function` which fits the model to the dataset and returns the
        log likelihood value defining how well the model fitted the data.

        It handles many other tasks, such as visualization, outputting results to hard-disk and storing results in
        a format that can be loaded after the model-fit is complete.

        This class is used for model-fits which fit a CTI model via a `CTI2D` object to a charge injection
        imaging dataset.

        Parameters
        ----------
        dataset
            The charge injection dataset that the model is fitted to.
        clocker
            The CTI arctic clocker used by the non-linear search and model-fit.
        settings_cti
            The settings controlling aspects of the CTI model in this model-fit.
        dataset_full
            The full dataset, which is visualized separate from the `dataset` that is fitted, which for example may
            not have the FPR masked and thus enable visualization of the FPR.
        """
        super().__init__()

        self.dataset = dataset
        self.clocker = clocker
        self.settings_cti = settings_cti
        self.dataset_full = dataset_full

        self.preloads = Preloads()

        parallel_fast_index_list = None
        parallel_fast_column_lists = None

        serial_fast_index_list = None
        serial_fast_row_lists = None

        if self.clocker.parallel_fast_mode and not self.clocker.serial_fast_mode:
            (
                parallel_fast_index_list,
                parallel_fast_column_lists,
            ) = clocker.fast_indexes_from(data=dataset.pre_cti_data, for_parallel=True)

        elif not self.clocker.parallel_fast_mode and self.clocker.serial_fast_mode:
            serial_fast_index_list, serial_fast_row_lists = clocker.fast_indexes_from(
                data=dataset.pre_cti_data, for_parallel=False
            )

        elif self.clocker.parallel_fast_mode and self.clocker.serial_fast_mode:
            raise exc.ClockerException(
                "Both parallel fast model and serial fast mode cannot be turned on.\n"
                "Only switch on parallel fast mode for parallel + serial clocking."
            )

        self.preloads = Preloads(
            parallel_fast_index_list=parallel_fast_index_list,
            parallel_fast_column_lists=parallel_fast_column_lists,
            serial_fast_index_list=serial_fast_index_list,
            serial_fast_row_lists=serial_fast_row_lists,
        )

    def region_list_from(self, model: af.Collection) -> List:
        """
        Inspects the CTI model and determines which regions are fitted for and therefore should be visualized.

        For example, if the model only includes parallel CTI, the serial regions are not fitted for and thus are not
        visualized.

        Parameters
        ----------
        model
            The CTI model, composed via PyAutoFit, which represents the parallel and serial CTI model compoenents
            fitted for by the non-linear search.

        Returns
        -------
        A list of the regions fitted for by the model and therefore visualized.

        """
        if model.cti.serial_ccd is None:
            return ["parallel_fpr", "parallel_eper"]
        elif model.cti.parallel_ccd is None:
            return ["serial_fpr", "serial_eper"]
        return ["parallel_fpr", "parallel_eper", "serial_fpr", "serial_eper"]

    def modify_before_fit(self, paths: af.DirectoryPaths, model: af.Collection):
        """
        This function is called immediately before the non-linear search begins and performs final tasks and checks
        before it begins.

        This function:

         1) Visualizes the charge injection imaging dataset, which does not change during the analysis and thus can be
            done once.

         2) Checks if the noise-map is fixed (it is not if hyper functionality is on), and if it is fixed it
            sets the noise-normalization to the preloads for computational speed.

        Parameters
        ----------
        paths
            The PyAutoFit paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization and the pickled objects used by the aggregator output by this function.
        model
            The PyAutoFit model object, which includes model components representing the galaxies that are fitted to
            the imaging data.
        """

        if paths.is_complete:
            return self

        if not model.has(HyperCINoiseCollection):
            noise_normalization = aa.util.fit.noise_normalization_with_mask_from(
                noise_map=self.dataset.noise_map, mask=self.dataset.mask
            )

            self.preloads.noise_normalization = noise_normalization

            logger.info(
                "PRELOADS - Noise Normalization preloaded for model-fit (noise-map is fixed)."
            )

        return self

    def log_likelihood_function(self, instance: af.ModelInstance) -> float:
        """
        Determine the fitness of a particular model

        Parameters
        ----------
        instance

        Returns
        -------
        fit: Fit
            How fit the model is and the model
        """

        self.settings_cti.check_total_density_within_range(
            parallel_traps=instance.cti.parallel_trap_list,
            serial_traps=instance.cti.serial_trap_list,
        )

        fit = self.fit_via_instance_and_dataset_from(
            instance=instance, dataset=self.dataset, hyper_noise_scale=True
        )

        return fit.figure_of_merit

    def fit_via_instance_and_dataset_from(
        self,
        instance: af.ModelInstance,
        dataset: ImagingCI,
        hyper_noise_scale: bool = True,
    ) -> FitImagingCI:
        hyper_noise_scalar_dict = None

        if hyper_noise_scale and hasattr(instance, "hyper_noise"):
            hyper_noise_scalar_dict = instance.hyper_noise.as_dict

        post_cti_data = self.clocker.add_cti(
            data=dataset.pre_cti_data, cti=instance.cti, preloads=self.preloads
        )

        return FitImagingCI(
            dataset=dataset,
            post_cti_data=post_cti_data,
            hyper_noise_scalar_dict=hyper_noise_scalar_dict,
            preloads=self.preloads,
        )

    def fit_via_instance_from(
        self, instance: af.ModelInstance, hyper_noise_scale: bool = True
    ) -> FitImagingCI:
        return self.fit_via_instance_and_dataset_from(
            instance=instance,
            dataset=self.dataset,
            hyper_noise_scale=hyper_noise_scale,
        )

    def save_attributes(self, paths: af.DirectoryPaths):
        """
        Before the model-fit via the non-linear search begins, this routine saves attributes of the `Analysis` object
        to the `pickles` folder such that they can be loaded after the analysis using PyAutoFit's database and
        aggregator tools.

        For this analysis the following are output:

        - The charge injection dataset (data / noise-map / pre cti data / cosmic ray map / layout / settings etc.).
        - The mask applied to the dataset.
        - The clocker used for modeling / clocking CTI.
        - The settings used for modeling / clocking CTI.
        - The full 1D dataset (e.g. unmasked, used for visualizariton).

        It is common for these attributes to be loaded by many of the template aggregator functions given in the
        `aggregator` modules. For example, when using the database tools to reperform a fit, this will by default
        load the dataset, settings and other attributes necessary to perform a fit using the attributes output by
        this function.

        Parameters
        ----------
        paths
            The PyAutoFit paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization,and the pickled objects used by the aggregator output by this function.
        """

        paths.save_json(
            name="clocker",
            object_dict=to_dict(self.clocker),
        )

        paths.save_json(
            name="settings_cti",
            object_dict=to_dict(self.settings_cti),
        )

        if conf.instance["visualize"]["plots"]["combined_only"]:
            return

        def output_dataset(dataset, prefix):
            paths.save_fits(
                name="data",
                hdu=dataset.data.hdu_for_output,
                prefix=prefix,
            )
            paths.save_fits(
                name="noise_map",
                hdu=dataset.noise_map.hdu_for_output,
                prefix=prefix,
            )
            paths.save_fits(
                name="pre_cti_data",
                hdu=dataset.pre_cti_data.hdu_for_output,
                prefix=prefix,
            )
            paths.save_json(
                name="layout",
                object_dict=to_dict(dataset.layout),
                prefix=prefix,
            )
            paths.save_fits(
                name="mask",
                hdu=dataset.mask.hdu_for_output,
                prefix=prefix,
            )

            if self.dataset.settings_dict is not None:
                paths.save_json(
                    name="settings_dict",
                    object_dict=self.dataset.settings_dict,
                    prefix="dataset",
                )

        output_dataset(dataset=self.dataset, prefix="dataset")

        if self.dataset_full is not None:
            output_dataset(dataset=self.dataset_full, prefix="dataset_full")

    def in_ascending_fpr_order_from(self, quantity_list, fpr_value_list):
        if not conf.instance["visualize"]["general"]["general"][
            "subplot_ascending_fpr"
        ]:
            return quantity_list

        indexes = sorted(range(len(fpr_value_list)), key=lambda k: fpr_value_list[k])

        return [quantity_list[i] for i in indexes]

    def visualize_before_fit(self, paths: af.DirectoryPaths, model: af.Collection):
        if conf.instance["visualize"]["plots"]["combined_only"]:
            return

        visualizer = VisualizerImagingCI(visualize_path=paths.image_path)

        region_list = self.region_list_from(model=model)

        if conf.instance["visualize"]["plots"]["dataset"]["fpr_non_uniformity"]:
            region_list += ["fpr_non_uniformity"]

        visualizer.visualize_dataset(dataset=self.dataset)
        visualizer.visualize_dataset_regions(
            dataset=self.dataset, region_list=region_list
        )

        if self.dataset_full is not None:
            visualizer.visualize_dataset(
                dataset=self.dataset_full, folder_suffix="_full"
            )
            visualizer.visualize_dataset_regions(
                dataset=self.dataset_full,
                region_list=region_list,
                folder_suffix="_full",
            )

    def visualize_before_fit_combined(
        self, analyses, paths: af.DirectoryPaths, model: af.Collection
    ):
        if analyses is None:
            return

        visualizer = VisualizerImagingCI(visualize_path=paths.image_path)

        region_list = self.region_list_from(model=model)

        if conf.instance["visualize"]["plots"]["dataset"]["fpr_non_uniformity"]:
            region_list += ["fpr_non_uniformity"]

        dataset_list = [analysis.dataset for analysis in analyses]
        fpr_value_list = [dataset.fpr_value for dataset in dataset_list]

        dataset_list = self.in_ascending_fpr_order_from(
            quantity_list=dataset_list,
            fpr_value_list=fpr_value_list,
        )

        visualizer.visualize_dataset_combined(
            dataset_list=dataset_list,
        )

        visualizer.visualize_dataset_regions_combined(
            dataset_list=dataset_list,
            region_list=region_list,
        )

        if self.dataset_full is not None:
            dataset_full_list = [analysis.dataset_full for analysis in analyses]

            dataset_full_list = self.in_ascending_fpr_order_from(
                quantity_list=dataset_full_list,
                fpr_value_list=fpr_value_list,
            )

            visualizer.visualize_dataset_combined(
                dataset_list=dataset_full_list,
                folder_suffix="_full",
                filename_suffix="_full",
            )
            visualizer.visualize_dataset_regions_combined(
                dataset_list=dataset_full_list,
                region_list=region_list,
                folder_suffix="_full",
                filename_suffix="_full",
            )

    def visualize(
        self,
        paths: af.DirectoryPaths,
        instance: af.ModelInstance,
        during_analysis: bool,
    ):
        if conf.instance["visualize"]["plots"]["combined_only"]:
            return

        fit = self.fit_via_instance_from(instance=instance)
        region_list = self.region_list_from(model=instance)

        visualizer = VisualizerImagingCI(visualize_path=paths.image_path)
        visualizer.visualize_fit(fit=fit, during_analysis=during_analysis)
        visualizer.visualize_fit_1d_regions(
            fit=fit, during_analysis=during_analysis, region_list=region_list
        )

        if self.dataset_full is not None:
            fit_full = self.fit_via_instance_and_dataset_from(
                instance=instance, dataset=self.dataset_full
            )

            visualizer.visualize_fit(
                fit=fit_full, during_analysis=during_analysis, folder_suffix="_full"
            )
            visualizer.visualize_fit_1d_regions(
                fit=fit_full,
                during_analysis=during_analysis,
                region_list=region_list,
                folder_suffix="_full",
            )

    def visualize_combined(
        self,
        analyses: List["AnalysisImagingCI"],
        paths: af.DirectoryPaths,
        instance: af.ModelInstance,
        during_analysis: bool,
    ):
        fit_list = [
            analysis.fit_via_instance_from(instance=instance) for analysis in analyses
        ]

        fpr_value_list = [fit.dataset.fpr_value for fit in fit_list]

        fit_list = self.in_ascending_fpr_order_from(
            quantity_list=fit_list,
            fpr_value_list=fpr_value_list,
        )

        region_list = self.region_list_from(model=instance)

        visualizer = VisualizerImagingCI(visualize_path=paths.image_path)
        visualizer.visualize_fit_combined(
            fit_list=fit_list, during_analysis=during_analysis
        )
        visualizer.visualize_fit_1d_regions_combined(
            fit_list=fit_list,
            region_list=region_list,
            during_analysis=during_analysis,
        )

        if self.dataset_full is not None:
            fit_full_list = [
                analysis.fit_via_instance_and_dataset_from(
                    instance=instance, dataset=analysis.dataset_full
                )
                for analysis in analyses
            ]

            fit_full_list = self.in_ascending_fpr_order_from(
                quantity_list=fit_full_list,
                fpr_value_list=fpr_value_list,
            )

            visualizer.visualize_fit_combined(
                fit_list=fit_full_list,
                during_analysis=during_analysis,
                folder_suffix="_full",
            )
            visualizer.visualize_fit_1d_regions_combined(
                fit_list=fit_full_list,
                region_list=region_list,
                during_analysis=during_analysis,
                folder_suffix="_full",
            )

    def make_result(
        self,
        samples: af.SamplesPDF,
    ) -> ResultImagingCI:
        return ResultImagingCI(samples=samples, analysis=self)



from typing import List, Optional

from arcticpy import CCD
from arcticpy import CCDPhase
from arcticpy import TrapInstantCapture

from autoconf.dictable import from_json, output_to_json

from autocti import exc


class AbstractClocker:
    def __init__(self, iterations: int = 1, verbosity: int = 0):
        """
        An abstract clocker, which wraps the c++ arctic CTI clocking algorithm in **PyAutoCTI**.

        Parameters
        ----------
        iterations
            The number of iterations used to correct CTI from an image.
        verbosity
            Whether to silence print statements and output from the c++ arctic call.
        """

        self.iterations = iterations
        self.verbosity = verbosity

    def ccd_from(self, ccd_phase: CCDPhase) -> CCD:
        """
        Returns a `CCD` object from a `CCDPhase` object.

        The `CCDPhase` describes the volume-filling behaviour of the CCD and is therefore used as a model-component
        in CTI calibration. To call arctic it needs converting to a `CCD` object.

        Parameters
        ----------
        ccd_phase
            The ccd phase describing the volume-filling behaviour of the CCD.

        Returns
        -------
        A `CCD` object based on the input phase which is passed to the c++ arctic.
        """
        if ccd_phase is not None:
            return CCD(phases=[ccd_phase], fraction_of_traps_per_phase=[1.0])

    def check_traps(
        self,
        trap_list_0: Optional[List[TrapInstantCapture]],
        trap_list_1: Optional[List[TrapInstantCapture]] = None,
    ):
        """
        Checks that there are trap species passed to the clocking algorithm and raises an exception if not.

        Parameters
        ----------
        trap_list
            A list of the trap species on the CCD which capture and release electrons during clock to as to add CTI.
        """
        if not any([trap_list_0, trap_list_1]):
            raise exc.ClockerException(
                "No Trap species were passed to the add_cti method"
            )

    def check_ccd(self, ccd_list: List[CCDPhase]):
        """
        Checks that there are trap species passed to the clocking algorithm and raises an exception if not.

        Parameters
        ----------
        ccd_list
            The ccd phase settings describing the volume-filling behaviour of the CCD which characterises the capture
            and release of electrons and therefore CTI.
        """
        if not any(ccd_list):
            raise exc.ClockerException("No CCD object was passed to the add_cti method")

    @classmethod
    def from_json(cls, file_path):
        return from_json(file_path=file_path)

    def output_to_json(self, file_path):
        output_to_json(obj=self, file_path=file_path)



import numpy as np
from typing import Callable

import autoarray.plot as aplt

from autoarray.fit.plot.fit_imaging_plotters import FitImagingPlotterMeta
from autoarray.plot.auto_labels import AutoLabels

from autocti.plot.abstract_plotters import Plotter
from autocti.charge_injection.fit import FitImagingCI


class FitImagingCIPlotter(Plotter):
    def __init__(
        self,
        fit: FitImagingCI,
        mat_plot_2d: aplt.MatPlot2D = aplt.MatPlot2D(),
        visuals_2d: aplt.Visuals2D = aplt.Visuals2D(),
        include_2d: aplt.Include2D = aplt.Include2D(),
        mat_plot_1d: aplt.MatPlot1D = aplt.MatPlot1D(),
        visuals_1d: aplt.Visuals1D = aplt.Visuals1D(),
        include_1d: aplt.Include1D = aplt.Include1D(),
        residuals_symmetric_cmap: bool = True,
    ):
        """
        Plots the attributes of `FitImagingCI` objects using the matplotlib methods `imshow()`, `plot()` and many other
        matplotlib functions which customize the plot's appearance.

        The `mat_plot_1d` and `mat_plot_2d` attribute wraps matplotlib function calls to make the figure. By default,
        the settings passed to every matplotlib function called are those specified in
        the `config/visualize/mat_wrap/*.ini` files, but a user can manually input values into `MatPlot1D` and
        `MatPlot2D` to customize the figure's appearance.

        Overlaid on the figure are visuals, contained in the `Visuals1D` and `Visuals2D` object. Attributes may be
        extracted from the `FitImagingCI` and plotted via the visuals object, if the corresponding entry
        is `True` in the `Include1D` and `Include2D` object or the `config/visualize/include.ini` file.

        Parameters
        ----------
        fit
            The fit to an imaging dataset the plotter plots.
        get_visuals_2d
            A function which extracts from the `FitImaging` the 2D visuals which are plotted on figures.
        mat_plot_2d
            Contains objects which wrap the matplotlib function calls that make the plot.
        visuals_2d
            Contains visuals that can be overlaid on the plot.
        include_2d
            Specifies which attributes of the `Array2D` are extracted and plotted as visuals.
        mat_plot_1d
            Contains objects which wrap the matplotlib function calls that make 1D plots.
        visuals_1d
            Contains 1D visuals that can be overlaid on 1D plots.
        include_1d
            Specifies which attributes of the `ImagingCI` are extracted and plotted as visuals for 1D plots.
        residuals_symmetric_cmap
            If true, the `residual_map` and `normalized_residual_map` are plotted with a symmetric color map such
            that `abs(vmin) = abs(vmax)`.
        """
        super().__init__(
            dataset=fit.dataset,
            mat_plot_2d=mat_plot_2d,
            include_2d=include_2d,
            visuals_2d=visuals_2d,
        )

        self.visuals_1d = visuals_1d
        self.include_1d = include_1d
        self.mat_plot_1d = mat_plot_1d

        self.fit = fit

        self._fit_imaging_meta_plotter = FitImagingPlotterMeta(
            fit=self.fit,
            get_visuals_2d=self.get_visuals_2d,
            mat_plot_2d=self.mat_plot_2d,
            include_2d=self.include_2d,
            visuals_2d=self.visuals_2d,
            residuals_symmetric_cmap=residuals_symmetric_cmap,
        )

        self.residuals_symmetric_cmap = residuals_symmetric_cmap

    def get_visuals_2d(self):
        return self.get_2d.via_fit_imaging_ci_from(fit=self.fit)

    @property
    def extract_region_from(self) -> Callable:
        return self.fit.dataset.layout.extract_region_from

    @property
    def extract_region_noise_map_from(self) -> Callable:
        return self.fit.dataset.layout.extract_region_noise_map_from

    def figures_2d(
        self,
        data: bool = False,
        noise_map: bool = False,
        signal_to_noise_map: bool = False,
        pre_cti_data: bool = False,
        post_cti_data: bool = False,
        residual_map: bool = False,
        normalized_residual_map: bool = False,
        chi_squared_map: bool = False,
    ):
        """
        Plots the individual attributes of the plotter's `FitImagingCI` object in 2D.

        The API is such that every plottable attribute of the `FitImagingCI` object is an input parameter of type bool
        of the function, which if switched to `True` means that it is plotted.

        Parameters
        ----------
        data
            Whether to make a 2D plot (via `imshow`) of the image data.
        noise_map
            Whether to make a 2D plot (via `imshow`) of the noise map.
        signal_to_noise_map
            Whether to make a 2D plot (via `imshow`) of the signal-to-noise map.
        pre_cti_data
            Whether to make a 2D plot (via `imshow`) of the pre-cti data.
        post_cti_data
            Whether to make a 2D plot (via `imshow`) of the post-cti data.
        residual_map
            Whether to make a 2D plot (via `imshow`) of the residual map.
        normalized_residual_map
            Whether to make a 2D plot (via `imshow`) of the normalized residual map.
        chi_squared_map
            Whether to make a 2D plot (via `imshow`) of the chi-squared map.
        """
        self._fit_imaging_meta_plotter.figures_2d(
            data=data,
            noise_map=noise_map,
            signal_to_noise_map=signal_to_noise_map,
            residual_map=residual_map,
            normalized_residual_map=normalized_residual_map,
            chi_squared_map=chi_squared_map,
        )

        if pre_cti_data:
            self.mat_plot_2d.plot_array(
                array=self.fit.pre_cti_data,
                visuals_2d=self.get_visuals_2d(),
                auto_labels=AutoLabels(title="Pre CTI Data", filename="pre_cti_data"),
            )

        if post_cti_data:
            self.mat_plot_2d.plot_array(
                array=self.fit.post_cti_data,
                visuals_2d=self.get_visuals_2d(),
                auto_labels=AutoLabels(
                    title="CI Post CTI Image", filename="post_cti_data"
                ),
            )

    def figures_1d(
        self,
        region,
        data: bool = False,
        data_logy: bool = False,
        noise_map: bool = False,
        signal_to_noise_map: bool = False,
        pre_cti_data: bool = False,
        post_cti_data: bool = False,
        residual_map: bool = False,
        residual_map_logy: bool = False,
        normalized_residual_map: bool = False,
        chi_squared_map: bool = False,
    ):
        """
        Plots the individual attributes of the plotter's `FitImagingCI` object in 1D.

        These 1D plots correspond to a region in 2D on the charge injection image, which is binned up over the parallel
        or serial direction to produce a 1D plot. For example, for the input `region=parallel_fpr`, this
        function extracts the FPR over each charge injection region and bins such that the 1D plot shows the FPR
        in the parallel direction.

        The API is such that every plottable attribute of the `FitImagingCI` object is an input parameter of type bool
        of the function, which if switched to `True` means that it is plotted.

        Parameters
        ----------
        region
            The region on the charge injection image where data is extracted and binned over the parallel or serial
            direction {"parallel_fpr", "parallel_eper", "serial_fpr", "serial_eper"}.
        data
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars and the model-fit overlaid.
        data_logy
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region, with the
            noise-map values included as error bars and the y-axis on a log10 scale and the model-fit overlaid.
        data
            Whether to make a 1D plot (via `plot`) of the image data extracted and binned over the region.
        noise_map
            Whether to make a 1D plot (via `plot`) of the noise-map extracted and binned over the region.
        signal_to_noise_map
            Whether to make a 1D plot (via `plot`) of the signal-to-noise map data extracted and binned over
            the region.
        pre_cti_data
            Whether to make a 1D plot (via `plot`) of the pre-cti data extracted and binned over the region.
        post_cti_data
            Whether to make a 1D plot (via `plot`) of the post-cti data extracted and binned over the
            region.
        residual_map
            Whether to make a 1D plot (via `plot`) of the residual map extracted and binned over the region, with the
            noise-map values included as error bars and the model-fit overlaid.
        residual_map_logy
            Whether to make a 1D plot (via `plot`) of the residual map extracted and binned over the region, with the
            noise-map values included as error bars and the y-axis on a log10 scale and the model-fit overlaid.
        normalized_residual_map
            Whether to make a 1D plot (via `plot`) of the normalized residual map extracted and binned over the
            region.
        chi_squared_map
            Whether to make a 1D plot (via `plot`) of the chi-squared map extracted and binned over the
            region.
        """

        title_str = self.title_str_from(region=region)

        y = self.extract_region_from(array=self.fit.data, region=region)
        y_errors = self.extract_region_noise_map_from(
            array=self.fit.noise_map, region=region
        )
        y_extra = self.extract_region_from(array=self.fit.model_data, region=region)

        should_plot_zero = self.should_plot_zero_from(region=region)

        if data:
            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar",
                y_errors=y_errors,
                y_extra=y_extra,
                should_plot_zero=should_plot_zero,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Data {title_str}",
                    yunit="e-",
                    filename=f"data_{region}",
                ),
            )

        if data_logy:
            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar_logy",
                y_errors=y_errors,
                y_extra=y_extra,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Data {title_str} [log10]",
                    yunit="e-",
                    filename=f"data_logy_{region}",
                ),
            )

        if noise_map:
            y = self.extract_region_from(array=self.fit.image, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Noise-Map {title_str}",
                    yunit="e-",
                    filename=f"noise_map_{region}",
                ),
            )

        if signal_to_noise_map:
            y = self.extract_region_from(
                array=self.fit.signal_to_noise_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Signal-To-Noise Map {title_str}",
                    yunit="",
                    filename=f"signal_to_noise_map_{region}",
                ),
            )

        if pre_cti_data:
            y = self.extract_region_from(array=self.fit.pre_cti_data, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"CI Pre CTI {title_str}",
                    yunit="e-",
                    filename=f"pre_cti_data_{region}",
                ),
            )

        if post_cti_data:
            y = self.extract_region_from(array=self.fit.post_cti_data, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                should_plot_zero=should_plot_zero,
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"CI Post CTI {title_str}",
                    yunit="e-",
                    filename=f"post_cti_data_{region}",
                ),
            )

        if residual_map:
            y = self.extract_region_from(array=self.fit.residual_map, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar",
                y_errors=y_errors,
                should_plot_zero=should_plot_zero,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Residual Map {title_str}",
                    yunit="e-",
                    filename=f"residual_map_{region}",
                ),
            )

        if residual_map_logy:
            y = self.extract_region_from(array=self.fit.residual_map, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                plot_axis_type_override="errorbar_logy",
                y_errors=y_errors,
                text_manual_dict=self.text_manual_dict_from(region=region),
                text_manual_dict_y=self.text_manual_dict_y_from(region=region),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Residual Map {title_str} [log10]",
                    yunit="e-",
                    filename=f"residual_map_logy_{region}",
                ),
            )

        if normalized_residual_map:
            y = self.extract_region_from(
                array=self.fit.normalized_residual_map, region=region
            )

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Normalized Residual Map {title_str}",
                    yunit=r"$\sigma$",
                    filename=f"normalized_residual_map_{region}",
                ),
            )

        if chi_squared_map:
            y = self.extract_region_from(array=self.fit.chi_squared_map, region=region)

            self.mat_plot_1d.plot_yx(
                y=y,
                x=range(len(y)),
                visuals_1d=self.visuals_1d,
                auto_labels=AutoLabels(
                    title=f"Chi-Squared Map {region}",
                    yunit=r"$\chi^2$",
                    filename=f"chi_squared_map_{region}",
                ),
            )

    def subplot(
        self,
        data: bool = False,
        noise_map: bool = False,
        signal_to_noise_map: bool = False,
        pre_cti_data: bool = False,
        post_cti_data: bool = False,
        residual_map: bool = False,
        normalized_residual_map: bool = False,
        chi_squared_map: bool = False,
        auto_filename: str = "subplot_fit",
    ):
        """
        Plots the individual attributes of the plotter's `FitImagingCI` object in 2D on a subplot.

        The API is such that every plottable attribute of the `FitImagingCI` object is an input parameter of type bool
        of the function, which if switched to `True` means that it is included on the subplot.

        Parameters
        ----------
        data
            Whether to include a 2D plot (via `imshow`) of the image data.
        noise_map
            Whether to include a 2D plot (via `imshow`) noise map.
        signal_to_noise_map
            Whether to include a 2D plot (via `imshow`) signal-to-noise map.
        pre_cti_data
            Whether to include a 2D plot (via `imshow`) of the pre-cti data.
        post_cti_data
            Whether to include a 2D plot (via `imshow`) of the post-cti data.
        residual_map
            Whether to include a 2D plot (via `imshow`) residual map.
        normalized_residual_map
            Whether to include a 2D plot (via `imshow`) normalized residual map.
        chi_squared_map
            Whether to include a 2D plot (via `imshow`) chi-squared map.
        auto_filename
            The default filename of the output subplot if written to hard-disk.
        """
        self._subplot_custom_plot(
            data=data,
            noise_map=noise_map,
            signal_to_noise_map=signal_to_noise_map,
            pre_cti_data=pre_cti_data,
            post_cti_data=post_cti_data,
            residual_map=residual_map,
            normalized_residual_map=normalized_residual_map,
            chi_squared_map=chi_squared_map,
            auto_labels=AutoLabels(filename=auto_filename),
        )

    def subplot_fit(self):
        """
        Standard subplot of the attributes of the plotter's `FitImaging` object.
        """
        return self.subplot(
            data=True,
            signal_to_noise_map=True,
            pre_cti_data=True,
            post_cti_data=True,
            normalized_residual_map=True,
            chi_squared_map=True,
        )

    def subplot_1d(self, region: str):
        """
        Plots the individual attributes of the plotter's `FitImagingCI` object in 1D on a subplot.

        These 1D plots correspond to a region in 2D on the charge injection image, which is binned up over the parallel
        or serial direction to produce a 1D plot. For example, for the input `region=parallel_fpr`, this
        function extracts the FPR over each charge injection region and bins such that the 1D plot shows the FPR
        in the parallel direction.

        The function plots the image, noise map, pre-cti data and signal to noise map in 1D on the subplot.

        Parameters
        ----------
        region
            The region on the charge injection image where data is extracted and binned over the parallel or serial
            direction {"parallel_fpr", "parallel_eper", "serial_fpr", "serial_eper"}
        """

        self.open_subplot_figure(number_subplots=6)

        self.figures_1d(data=True, region=region)
        self.figures_1d(signal_to_noise_map=True, region=region)
        self.figures_1d(pre_cti_data=True, region=region)
        self.figures_1d(post_cti_data=True, region=region)
        self.figures_1d(normalized_residual_map=True, region=region)
        self.figures_1d(chi_squared_map=True, region=region)

        self.mat_plot_1d.output.subplot_to_figure(
            auto_filename=f"subplot_1d_fit_ci_{region}"
        )
        self.close_subplot_figure()

    def subplot_noise_scaling_map_dict(self):
        """
        Plots the noise-scaling maps of the plotter's `FitImagingCI` object in 2D on a subplot.
        """

        self.open_subplot_figure(number_subplots=len(self.fit.noise_scaling_map_dict))

        for key, noise_scaling_map in self.fit.noise_scaling_map_dict.items():
            self.mat_plot_2d.plot_array(
                array=noise_scaling_map,
                visuals_2d=self.get_visuals_2d(),
                auto_labels=AutoLabels(title=f"Noise Scaling Map {key}"),
            )

        self.mat_plot_2d.output.subplot_to_figure(
            auto_filename=f"subplot_noise_scaling_map_dict"
        )
        self.mat_plot_2d.figure.close()



from typing import Optional, Tuple

from autocti.extract.one_d.overscan import Extract1DOverscan
from autocti.extract.one_d.fpr import Extract1DFPR
from autocti.extract.one_d.eper import Extract1DEPER
from autocti.extract.settings import SettingsExtract

import autoarray as aa


class Extract1DMaster:
    def __init__(
        self,
        region_list,
        shape_1d: Optional[Tuple[int]] = None,
        prescan: Optional[aa.type.Region1DLike] = None,
        overscan: Optional[aa.type.Region1DLike] = None,
    ):
        """
        Abstract class containing methods for extracting regions from a 1D line dataset which contains some sort of
        original signal whose profile before CTI is known (e.g. warm pixel, charge injection).

        This uses the `region_list`, which contains the signal's regions in pixel coordinates (x0, x1).

        Parameters
        ----------
        region_list
            Integer pixel coordinates specifying the corners of signal (x0, x1).
        prescan
            Integer pixel coordinates specifying the corners of the prescan (x0, x1).
        overscan
            Integer pixel coordinates specifying the corners of the overscan (x0, x1).
        """

        self.region_list = (
            list(map(aa.Region1D, region_list)) if region_list is not None else None
        )

        self.shape_1d = shape_1d

        self._prescan = prescan
        self._overscan = overscan

    @property
    def fpr(self):
        return Extract1DFPR(
            shape_1d=self.shape_1d,
            region_list=self.region_list,
            prescan=self._prescan,
            overscan=self._overscan,
        )

    @property
    def eper(self):
        return Extract1DEPER(
            shape_1d=self.shape_1d,
            region_list=self.region_list,
            prescan=self._prescan,
            overscan=self._overscan,
        )

    @property
    def overscan(self):
        return Extract1DOverscan(overscan=self._overscan)

    def regions_array_1d_from(self, array: aa.Array1D) -> aa.Array1D:
        """
        Extract all of the charge-injection regions from an input `array1D` object and returns them as a new `array1D`
        object where these extracted regions are included and all other entries are zeros.

        The diagram below illustrates the regions that are extracted from the input array:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCDPhase panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sss]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
          [...][ccccccccccccccccccccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted array keeps just the charge injection region and replaces all other values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][000000000000000000000][000]
          [000][ccccccccccccccccccccc][000]
        | [000][ccccccccccccccccccccc][000]    |
        | [000][000000000000000000000][000]    | Direction
        P [000][000000000000000000000][000]    | of
        | [000][ccccccccccccccccccccc][000]    | clocking
          [000][ccccccccccccccccccccc][000]    |

        []     [=====================]
               <---------S----------

        """

        array_1d_of_regions = array.native.copy() * 0.0

        for region in self.region_list:
            array_1d_of_regions[region.slice] += array.native[region.slice]

        return array_1d_of_regions

    def non_regions_array_1d_from(self, array: aa.Array1D) -> aa.Array1D:
        """
        Extract all of the data values in an input `array1D` that do not overlap the charge injection regions. This
        includes many areas of the image (e.g. the serial prescan, serial overscan) but is typically used to extract
        a `array1D` that contains the parallel trails that follow the charge-injection regions.

        The diagram below illustrates the `array1D` that is extracted from the input array:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCDPhase panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [tptpptptptpptpptpptpt]
               [tptptptpptpttptptptpt]
          [...][ttttttttttttttttttttt][sss]
          [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
          [...][ccccccccccccccccccccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted array keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [tptpptptptpptpptpptpt]
               [tptptptpptpttptptptpt]
          [000][ttttttttttttttttttttt][000]
          [000][000000000000000000000][000]
        | [000][000000000000000000000][000]    |
        | [000][ttttttttttttttttttttt][000]    | Direction
        P [000][ttttttttttttttttttttt][000]    | of
        | [000][000000000000000000000][000]    | clocking
          [000][000000000000000000000][000]    |

        []     [=====================]
               <---------S----------
        """

        array_1d_non_regions_ci = array.native.copy()

        for region in self.region_list:
            array_1d_non_regions_ci[region.slice] = 0.0

        return array_1d_non_regions_ci

    def array_1d_of_edges_and_eper_from(
        self,
        array: aa.Array1D,
        fpr_pixels: Tuple[int, int] = None,
        eper_pixels: Tuple[int, int] = None,
    ) -> aa.Array1D:
        """
            Extract all of the data values in an input `array1D` corresponding to the parallel front edges and trails of
            each the charge-injection region.

            One can specify the range of rows that are extracted, for example:

            fpr_pixels = (0, 1) will extract just the first leading front edge row.
            fpr_pixels = (0, 2) will extract the leading two front edge rows.
            trails_pixels = (0, 1) will extract the first row of trails closest to the charge injection region.

            The diagram below illustrates the arrays that are extracted from the input array for `fpr_pixels=(0,1)`
            and `trails_pixels=(0,1)`:

            ---KEY---
            ---------

            [] = read-out electronics   [==========] = read-out register

            [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
            [xxxxxxxxxx] = CCDPhase panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
            [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

            P = Parallel Direction      S = Serial Direction

                   [tptpptptptpptpptpptpt]
                   [tptptptpptpttptptptpt]
              [...][ttttttttttttttttttttt][sss]
              [...][ccccccccccccccccccccc][sss]
            | [...][ccccccccccccccccccccc][sss]    |
            | [...][ttttttttttttttttttttt][sss]    | Direction
            P [...][ttttttttttttttttttttt][sss]    | of
            | [...][ccccccccccccccccccccc][sss]    | clocking
              [...][ccccccccccccccccccccc][sss]    |

            []     [=====================]
                   <---------S----------

            The extracted array keeps just the leading edges and trails following all charge injection scans and
            replaces all other values with 0s:

                   [000000000000000000000]
                   [000000000000000000000]
              [000][ttttttttttttttttttttt][000]
              [000][000000000000000000000][000]
            | [000][ccccccccccccccccccccc][000]    |
            | [000][000000000000000000000][000]    | Direction
            P [000][ttttttttttttttttttttt][000]    | of
            | [000][000000000000000000000][000]    | clocking
              [000][ccccccccccccccccccccc][000]    |

            []     [=====================]
                   <---------S----------

            Parameters
        ----------
            fpr_pixels
                The row indexes to extract the front edge between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows).
            trails_pixels
                The row indexes to extract the trails between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows)
        """
        array_1d_of_edges_and_eper = array.native.copy() * 0.0

        if fpr_pixels is not None:
            array_1d_of_edges_and_eper = self.fpr.add_to_array(
                new_array=array_1d_of_edges_and_eper,
                array=array,
                settings=SettingsExtract(pixels=fpr_pixels),
            )

        if eper_pixels is not None:
            array_1d_of_edges_and_eper = self.eper.add_to_array(
                new_array=array_1d_of_edges_and_eper,
                array=array,
                settings=SettingsExtract(pixels=eper_pixels),
            )

        return array_1d_of_edges_and_eper



from typing import List, Optional, Tuple

import autoarray as aa

from autocti.extract.two_d.parallel.eper import Extract2DParallelEPER
from autocti.extract.two_d.serial.abstract import Extract2DSerial
from autocti.extract.settings import SettingsExtract


class Extract2DSerialOverscanNoEPER(Extract2DSerial):
    def region_list_from(self, settings: SettingsExtract) -> List[aa.Region2D]:
        """
        Returns a list of the 2D serial overscan regions without EPER trails, between two input `pixels` indexes.

        (top-row, bottom-row, left-column, right-column) = (y0, y1, x0, x1)

        The serial overscan spans all columns of the image, thus the coordinates x0 and x1 do not change. y0 and y1
        are updated based on where serial EPER trails are not and the `pixels` input.

        Negative pixel values can be input into the `pixels` tuple, whereby columns in front of the serial overscan are
        also extracted.

        The diagram below illustrates the extraction for `pixels=(0, 1)`:

        [] = read-out electronics
        [==========] = read-out register
        [..........] = serial prescan
        [pppppppppp] = serial overscan
        [ssssssssss] = serial overscan
        [snsnsnsnsn] = serial overscan with no serial EPERs
        [f#ff#f#f#f] = signal region (FPR) (0 / 1 indicate the region index)
        [tttttttttt] = serial / serial charge injection region trail

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
           [...][ttttttttttttttttttttt][sns]
           [...][c1c1cc1c1cc1cc1ccc1cc][sns]
        |  [...][1c1c1cc1c1cc1ccc1cc1][sss]    |
        |  [...][ttttttttttttttttttttt][sns]    | Direction
        Par [...][ttttttttttttttttttttt][sns]    | of
        |  [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
        |/  [...][cc0ccc0cccc0cccc0cccc][sss]    \/

        []     [=====================]
               <---------Ser--------

        The extracted regions correspond to the serial overscan with no serial EPERs [sns] regions.

        For `pixels=(0,1)` the extracted arrays returned via the `array_2d_list_from()` function keep the first
        serial pixels across the columns of the overscan:

        array_2d_list[0] = [sns]

        Parameters
        ----------
        settings
            The settings used to extract the serial overscan region without EPERs from, which for example include
            the `pixels`  tuple specifying the range of pixel columns they are extracted between.
        """

        pixels = settings.pixels

        if settings.pixels_from_end is not None:
            pixels = (
                self.serial_overscan.total_columns - settings.pixels_from_end,
                self.serial_overscan.total_columns,
            )

        extract_parallel_eper = Extract2DParallelEPER(
            shape_2d=self.shape_2d,
            region_list=self.region_list,
            parallel_overscan=self.parallel_overscan,
            serial_overscan=self.serial_overscan,
            serial_prescan=self.serial_prescan,
        )

        parallel_eper_region_list = extract_parallel_eper.region_list_from(
            settings=SettingsExtract(pixels_from_end=-1)
        )

        region_list = []

        for i, region in enumerate(self.region_list):
            region_list.append(
                aa.Region2D(
                    (
                        parallel_eper_region_list[i].y0,
                        parallel_eper_region_list[i].y1,
                        self.serial_overscan.x0 + pixels[0],
                        self.serial_overscan.x0 + pixels[1],
                    )
                )
            )

        return region_list



import numpy as np
from typing import Optional

import autoarray as aa

from autoarray.dataset.imaging.simulator import SimulatorImaging
from autoarray.dataset import preprocess

from autocti.dataset_1d.dataset_1d.dataset_1d import Dataset1D
from autocti.extract.settings import SettingsExtract
from autocti.layout.one_d import Layout1D
from autocti.clocker.one_d import Clocker1D
from autocti.model.model_util import CTI1D


class SimulatorDataset1D(SimulatorImaging):
    def __init__(
        self,
        pixel_scales: aa.type.PixelScales,
        norm: float,
        read_noise: Optional[float] = None,
        add_poisson_noise: bool = False,
        charge_noise: Optional[float] = None,
        noise_if_add_noise_false: float = 0.1,
        noise_seed: int = -1,
    ):
        """
        A class representing a Imaging observation, using the shape of the data, the pixel scale,
        psf, exposure time, etc.

        Parameters
        ----------
        exposure_time_map
            The exposure time of an observation using this data_type.
        """

        super().__init__(
            read_noise=read_noise,
            exposure_time=1.0,
            add_poisson_noise=add_poisson_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

        self.pixel_scales = pixel_scales
        self.norm = norm
        self.charge_noise = charge_noise

    def pre_cti_data_from(
        self, layout: Layout1D, pixel_scales: aa.type.PixelScales
    ) -> aa.Array1D:
        """Use this charge injection layout_ci to generate a pre-cti charge injection image. This is performed by \
        going to its charge injection regions and adding the charge injection normalization value.

        Parameters
        ----------
        shape_native
            The image_shape of the pre_cti_datas to be created.
        """

        pre_cti_data = np.zeros(layout.shape_1d)

        for region in layout.region_list:
            pre_cti_data[region.slice] += self.norm

        return aa.Array1D.no_mask(values=pre_cti_data, pixel_scales=pixel_scales)

    def via_layout_from(
        self, layout: Layout1D, clocker: Clocker1D, cti: CTI1D
    ) -> Dataset1D:
        """Simulate a charge injection data, including effects like noises.

        Parameters
        ----------
        pre_cti_data
        cosmic_ray_map
            The dimensions of the output simulated charge injection data.
        frame_geometry
            The quadrant geometry of the simulated data, defining where the parallel / serial overscans are and \
            therefore the direction of clocking and rotations before input into the cti algorithm.
        layout : layout_ci.Layout1DSimulate
            The charge injection layout_ci (regions, normalization, etc.) of the charge injection data.
        cti_params
            The CTI model parameters (trap density, trap release_timescales etc.).
        clocker
            The settings that control the cti clocking algorithm (e.g. ccd well_depth express option).
        read_noise
            The FWHM of the Gaussian read-noises added to the data.
        noise_seed
            Seed for the read-noises added to the data.
        """

        pre_cti_data = self.pre_cti_data_from(
            layout=layout, pixel_scales=self.pixel_scales
        )

        return self.via_pre_cti_data_from(
            pre_cti_data=pre_cti_data.native, layout=layout, clocker=clocker, cti=cti
        )

    def via_pre_cti_data_from(
        self, pre_cti_data: aa.Array1D, layout: Layout1D, clocker: Clocker1D, cti: CTI1D
    ) -> Dataset1D:
        if self.charge_noise is not None:
            pre_cti_data = layout.extract.fpr.add_gaussian_noise_to(
                array=pre_cti_data,
                noise_sigma=self.charge_noise,
                noise_seed=self.noise_seed,
                settings=SettingsExtract(
                    pixels_from_end=layout.extract.fpr.total_pixels_min
                ),
            )

        post_cti_data = clocker.add_cti(data=pre_cti_data.native, cti=cti)

        return self.via_post_cti_data_from(
            post_cti_data=post_cti_data, pre_cti_data=pre_cti_data, layout=layout
        )

    def via_post_cti_data_from(
        self, post_cti_data: aa.Array1D, pre_cti_data: aa.Array1D, layout: Layout1D
    ) -> Dataset1D:
        if self.read_noise is not None:
            data = preprocess.data_with_gaussian_noise_added(
                data=post_cti_data, sigma=self.read_noise, seed=self.noise_seed
            )
            noise_map = (
                self.read_noise
                * aa.Array1D.ones(
                    shape_native=layout.shape_1d, pixel_scales=self.pixel_scales
                ).native
            )
        else:
            data = post_cti_data
            noise_map = aa.Array1D.full(
                fill_value=self.noise_if_add_noise_false,
                shape_native=layout.shape_1d,
                pixel_scales=self.pixel_scales,
            ).native

        return Dataset1D(
            data=aa.Array1D.no_mask(values=data.native, pixel_scales=self.pixel_scales),
            noise_map=aa.Array1D.no_mask(
                values=noise_map, pixel_scales=self.pixel_scales
            ),
            pre_cti_data=aa.Array1D.no_mask(
                values=pre_cti_data.native, pixel_scales=self.pixel_scales
            ),
            layout=layout,
        )



import logging

import autoarray.plot as aplt

import autocti.plot as aplt

from autocti.model.visualizer import Visualizer
from autocti.model.visualizer import plot_setting

from autocti import exc

logger = logging.getLogger(__name__)

logger.setLevel(level="INFO")


class VisualizerImagingCI(Visualizer):
    def visualize_dataset(self, dataset, folder_suffix: str = ""):
        def should_plot(name):
            return plot_setting(section="dataset", name=name)

        mat_plot = self.mat_plot_2d_from(subfolders=f"dataset{folder_suffix}")

        dataset_plotter = aplt.ImagingCIPlotter(
            dataset=dataset, mat_plot_2d=mat_plot, include_2d=self.include_2d
        )

        if should_plot("subplot_dataset"):
            dataset_plotter.subplot_dataset()

        dataset_plotter.figures_2d(
            data=should_plot("data"),
            noise_map=should_plot("noise_map"),
            signal_to_noise_map=should_plot("signal_to_noise_map"),
            pre_cti_data=should_plot("pre_cti_data"),
            cosmic_ray_map=should_plot("cosmic_ray_map"),
        )

    def visualize_dataset_regions(self, dataset, region_list, folder_suffix: str = ""):
        def should_plot(name):
            return plot_setting(section="dataset", name=name)

        mat_plot = self.mat_plot_1d_from(subfolders=f"dataset{folder_suffix}")

        dataset_plotter = aplt.ImagingCIPlotter(
            dataset=dataset, mat_plot_1d=mat_plot, include_2d=self.include_2d
        )

        for region in region_list:
            try:
                if should_plot("subplot_dataset"):
                    dataset_plotter.subplot_1d(region=region)

                dataset_plotter.figures_1d(
                    region=region,
                    data=should_plot("data"),
                    data_logy=should_plot("data_logy"),
                    noise_map=should_plot("noise_map"),
                    signal_to_noise_map=should_plot("signal_to_noise_map"),
                    pre_cti_data=should_plot("pre_cti_data"),
                )

            except (exc.RegionException, TypeError, ValueError):
                logger.info(
                    f"VISUALIZATION - Could not visualize the ImagingCI 1D {region}"
                )

    def visualize_dataset_combined(
        self, dataset_list, folder_suffix: str = "", filename_suffix: str = ""
    ):
        def should_plot(name):
            return plot_setting(section="dataset", name=name)

        mat_plot = self.mat_plot_2d_from(subfolders=f"dataset_combined{folder_suffix}")

        dataset_plotter_list = [
            aplt.ImagingCIPlotter(
                dataset=dataset, mat_plot_2d=mat_plot, include_2d=self.include_2d
            )
            for dataset in dataset_list
        ]
        multi_plotter = aplt.MultiFigurePlotter(plotter_list=dataset_plotter_list)

        if should_plot("subplot_dataset"):
            multi_plotter.subplot_of_figure(
                func_name="figures_2d",
                figure_name="data",
                filename_suffix=filename_suffix,
            )
            multi_plotter.subplot_of_figure(
                func_name="figures_2d",
                figure_name="noise_map",
                filename_suffix=filename_suffix,
            )
            multi_plotter.subplot_of_figure(
                func_name="figures_2d",
                figure_name="signal_to_noise_map",
                filename_suffix=filename_suffix,
            )
            multi_plotter.subplot_of_figure(
                func_name="figures_2d",
                figure_name="pre_cti_data",
                filename_suffix=filename_suffix,
            )
            multi_plotter.subplot_of_figure(
                func_name="figures_2d",
                figure_name="pre_cti_data_residual_map",
                filename_suffix=filename_suffix,
            )
            multi_plotter.subplot_of_figure(
                func_name="figures_2d",
                figure_name="cosmic_ray_map",
                filename_suffix=filename_suffix,
            )

    def visualize_dataset_regions_combined(
        self,
        dataset_list,
        region_list,
        folder_suffix: str = "",
        filename_suffix: str = "",
    ):
        def should_plot(name):
            return plot_setting(section="dataset", name=name)

        mat_plot_1d = self.mat_plot_1d_from(
            subfolders=f"dataset_combined{folder_suffix}"
        )

        dataset_plotter_list = [
            aplt.ImagingCIPlotter(
                dataset=dataset, mat_plot_2d=mat_plot_1d, include_1d=self.include_1d
            )
            for dataset in dataset_list
        ]
        multi_plotter = aplt.MultiFigurePlotter(plotter_list=dataset_plotter_list)

        for region in region_list:
            try:
                if should_plot("data"):
                    multi_plotter.subplot_of_figure(
                        func_name="figures_1d",
                        figure_name="data",
                        region=region,
                        filename_suffix=f"{filename_suffix}_{region}",
                    )

                if should_plot("data_logy"):
                    multi_plotter.subplot_of_figure(
                        func_name="figures_1d",
                        figure_name="data_logy",
                        region=region,
                        filename_suffix=f"{filename_suffix}_{region}",
                    )

            except (exc.RegionException, TypeError, ValueError):
                logger.info(
                    f"VISUALIZATION - Could not visualize the ImagingCI 1D {region}"
                )

    def visualize_fit(self, fit, during_analysis, folder_suffix: str = ""):
        def should_plot(name):
            return plot_setting(section="fit", name=name)

        mat_plot = self.mat_plot_2d_from(subfolders=f"fit_dataset{folder_suffix}")

        fit_plotter = aplt.FitImagingCIPlotter(
            fit=fit, mat_plot_2d=mat_plot, include_2d=self.include_2d
        )

        fit_plotter.figures_2d(
            data=should_plot("data"),
            noise_map=should_plot("noise_map"),
            signal_to_noise_map=should_plot("signal_to_noise_map"),
            pre_cti_data=should_plot("pre_cti_data"),
            post_cti_data=should_plot("post_cti_data"),
            residual_map=should_plot("residual_map"),
            normalized_residual_map=should_plot("normalized_residual_map"),
            chi_squared_map=should_plot("chi_squared_map"),
        )

        if not during_analysis:
            if should_plot("all_at_end_png"):
                fit_plotter.figures_2d(
                    data=True,
                    noise_map=True,
                    signal_to_noise_map=True,
                    pre_cti_data=True,
                    post_cti_data=True,
                    residual_map=True,
                    normalized_residual_map=True,
                    chi_squared_map=True,
                )

            if should_plot("all_at_end_fits"):
                self.visualize_fit_in_fits(fit=fit)

        if should_plot("subplot_fit"):
            fit_plotter.subplot_fit()

    def visualize_fit_1d_regions(
        self, fit, region_list, during_analysis, folder_suffix: str = ""
    ):
        def should_plot(name):
            return plot_setting(section="fit", name=name)

        mat_plot = self.mat_plot_1d_from(subfolders=f"fit_dataset{folder_suffix}")

        fit_plotter = aplt.FitImagingCIPlotter(
            fit=fit, mat_plot_1d=mat_plot, include_1d=self.include_1d
        )

        for region in region_list:
            try:
                if should_plot("subplot_fit"):
                    fit_plotter.subplot_1d(region=region)

                fit_plotter.figures_1d(
                    region=region,
                    data=should_plot("data"),
                    data_logy=should_plot("data_logy"),
                    noise_map=should_plot("noise_map"),
                    signal_to_noise_map=should_plot("signal_to_noise_map"),
                    pre_cti_data=should_plot("pre_cti_data"),
                    post_cti_data=should_plot("post_cti_data"),
                    residual_map=should_plot("residual_map"),
                    residual_map_logy=should_plot("residual_map_logy"),
                    normalized_residual_map=should_plot("normalized_residual_map"),
                    chi_squared_map=should_plot("chi_squared_map"),
                )

                if not during_analysis:
                    if should_plot("all_at_end_png"):
                        fit_plotter.figures_1d(
                            region=region,
                            data=True,
                            noise_map=True,
                            signal_to_noise_map=True,
                            pre_cti_data=True,
                            post_cti_data=True,
                            residual_map=True,
                            residual_map_logy=True,
                            normalized_residual_map=True,
                            chi_squared_map=True,
                        )

            except (exc.RegionException, TypeError, ValueError):
                logger.info(
                    f"VISUALIZATION - Could not visualize the ImagingCI 1D {region}"
                )

    def visualize_fit_combined(
        self, fit_list, during_analysis, folder_suffix: str = ""
    ):
        def should_plot(name):
            return plot_setting(section="fit", name=name)

        mat_plot = self.mat_plot_2d_from(
            subfolders=f"fit_dataset_combined{folder_suffix}"
        )

        fit_plotter_list = [
            aplt.FitImagingCIPlotter(
                fit=fit, mat_plot_2d=mat_plot, include_2d=self.include_2d
            )
            for fit in fit_list
        ]
        multi_plotter = aplt.MultiFigurePlotter(plotter_list=fit_plotter_list)

        if should_plot("residual_map"):
            multi_plotter.subplot_of_figure(
                func_name="figures_2d", figure_name="residual_map"
            )

        if should_plot("normalized_residual_map"):
            multi_plotter.subplot_of_figure(
                func_name="figures_2d", figure_name="normalized_residual_map"
            )

        if should_plot("chi_squared_map"):
            multi_plotter.subplot_of_figure(
                func_name="figures_2d", figure_name="chi_squared_map"
            )

    def visualize_fit_1d_regions_combined(
        self, fit_list, region_list, during_analysis, folder_suffix: str = ""
    ):
        def should_plot(name):
            return plot_setting(section="fit", name=name)

        mat_plot = self.mat_plot_1d_from(
            subfolders=f"fit_dataset_combined{folder_suffix}"
        )

        fit_plotter_list = [
            aplt.FitImagingCIPlotter(
                fit=fit, mat_plot_1d=mat_plot, include_1d=self.include_1d
            )
            for fit in fit_list
        ]
        multi_plotter = aplt.MultiFigurePlotter(plotter_list=fit_plotter_list)

        for region in region_list:
            try:
                if should_plot("data"):
                    multi_plotter.subplot_of_figure(
                        func_name="figures_1d",
                        figure_name="data",
                        region=region,
                        filename_suffix=f"_{region}",
                    )

                if should_plot("data_logy"):
                    multi_plotter.subplot_of_figure(
                        func_name="figures_1d",
                        figure_name="data_logy",
                        region=region,
                        filename_suffix=f"_{region}",
                    )

                if should_plot("residual_map"):
                    multi_plotter.subplot_of_figure(
                        func_name="figures_1d",
                        figure_name="residual_map",
                        region=region,
                        filename_suffix=f"_{region}",
                    )

                if should_plot("residual_map_logy"):
                    multi_plotter.subplot_of_figure(
                        func_name="figures_1d",
                        figure_name="residual_map_logy",
                        region=region,
                        filename_suffix=f"_{region}",
                    )

                if should_plot("normalized_residual_map"):
                    multi_plotter.subplot_of_figure(
                        func_name="figures_1d",
                        figure_name="normalized_residual_map",
                        region=region,
                        filename_suffix=f"_{region}",
                    )

                if should_plot("chi_squared_map"):
                    multi_plotter.subplot_of_figure(
                        func_name="figures_1d",
                        figure_name="chi_squared_map",
                        region=region,
                        filename_suffix=f"_{region}",
                    )

            except (exc.RegionException, TypeError, ValueError):
                logger.info(
                    f"VISUALIZATION - Could not visualize the ImagingCI 1D {region}"
                )

    def visualize_fit_in_fits(self, fit):
        mat_plot_2d = self.mat_plot_2d_from(subfolders="fit_imaging/fit", format="fit")

        fit_plotter = aplt.FitImagingCIPlotter(
            fit=fit, mat_plot_2d=mat_plot_2d, include_2d=self.include_2d
        )

        fit_plotter.figures_2d(
            data=True,
            noise_map=True,
            signal_to_noise_map=True,
            pre_cti_data=True,
            post_cti_data=True,
            residual_map=True,
            normalized_residual_map=True,
            chi_squared_map=True,
        )



import numpy as np
from typing import Tuple

import autoarray as aa

from autoarray import exc

from autocti.extract.settings import SettingsExtract
from autocti.layout.two_d import Layout2D


class SettingsMask2D:
    def __init__(
        self,
        parallel_fpr_pixels: Tuple[int, int] = None,
        parallel_eper_pixels: Tuple[int, int] = None,
        serial_fpr_pixels: Tuple[int, int] = None,
        serial_eper_pixels: Tuple[int, int] = None,
        cosmic_ray_parallel_buffer: int = 10,
        cosmic_ray_serial_buffer: int = 10,
        cosmic_ray_diagonal_buffer: int = 3,
    ):
        self.parallel_fpr_pixels = parallel_fpr_pixels
        self.parallel_eper_pixels = parallel_eper_pixels
        self.serial_fpr_pixels = serial_fpr_pixels
        self.serial_eper_pixels = serial_eper_pixels

        self.cosmic_ray_parallel_buffer = cosmic_ray_parallel_buffer
        self.cosmic_ray_serial_buffer = cosmic_ray_serial_buffer
        self.cosmic_ray_diagonal_buffer = cosmic_ray_diagonal_buffer


class Mask2D(aa.Mask2D):
    @classmethod
    def manual(cls, mask, pixel_scales, origin=(0.0, 0.0), invert=False):
        """
        Create a Mask2D (see *Mask2D.__new__*) by inputting the array values in 2D, for example:

        mask=np.array([[False, False],
                       [True, False]])

        mask=[[False, False],
               [True, False]]

        Parameters
        ----------
        mask
            The bool values of the mask input as an ndarray of shape [total_y_pixels, total_x_pixels ]or a list of
            lists.
        pixel_scales
            The (y,x) arcsecond-to-pixel units conversion factor of every pixel. If this is input as a `float`,
            it is converted to a (float, float).
        origin : (float, float)
            The (y,x) scaled units origin of the mask's coordinate system.
        invert
            If `True`, the ``bool``'s of the input ``mask`` are inverted, for example `False`'s become `True`
            and visa versa.
        """
        if type(mask) is list:
            mask = np.asarray(mask).astype("bool")

        if invert:
            mask = np.invert(mask)

        pixel_scales = aa.util.geometry.convert_pixel_scales_2d(
            pixel_scales=pixel_scales
        )

        if len(mask.shape) != 2:
            raise exc.MaskException("The input mask is not a two dimensional array")

        return cls(mask=mask, pixel_scales=pixel_scales, origin=origin)

    @classmethod
    def all_false(cls, shape_native, pixel_scales, origin=(0.0, 0.0), invert=False):
        """Create a mask where all pixels are `False` and therefore unmasked.

        Parameters
        ----------
        mask
            The bool values of the mask input as an ndarray of shape [total_y_pixels, total_x_pixels ]or a list of
            lists.
        pixel_scales
            The (y,x) arcsecond-to-pixel units conversion factor of every pixel. If this is input as a `float`,
            it is converted to a (float, float).
        origin : (float, float)
            The (y,x) scaled units origin of the mask's coordinate system.
        invert
            If `True`, the ``bool``'s of the input ``mask`` are inverted, for example `False`'s become `True`
            and visa versa.
        """
        return cls.manual(
            mask=np.full(shape=shape_native, fill_value=False),
            pixel_scales=pixel_scales,
            origin=origin,
            invert=invert,
        )

    @classmethod
    def from_masked_regions(cls, shape_native, pixel_scales, masked_regions):
        mask = cls.all_false(shape_native=shape_native, pixel_scales=pixel_scales)
        masked_regions = list(
            map(lambda region: aa.Region2D(region=region), masked_regions)
        )
        for region in masked_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        return mask

    @classmethod
    def from_cosmic_ray_map_buffed(cls, cosmic_ray_map, settings=SettingsMask2D()):
        """
        Returns the mask used for CTI Calibration, which is all `False` unless specific regions are input for masking.

        Parameters
        ----------
        cosmic_ray_map : array_2d.Array2D
            2D arrays flagging where cosmic rays on the image.
        cosmic_ray_parallel_buffer
            The number of pixels from each ray pixels are masked in the parallel direction.
        cosmic_ray_serial_buffer
            The number of pixels from each ray pixels are masked in the serial direction.
        cosmic_ray_diagonal_buffer
            The number of pixels from each ray pixels are masked in the digonal up from the parallel + serial direction.
        """
        mask = cls.all_false(
            shape_native=cosmic_ray_map.shape_native,
            pixel_scales=cosmic_ray_map.pixel_scales,
        )

        cosmic_ray_mask = (cosmic_ray_map.native > 0.0).astype("bool")

        for y in range(mask.shape[0]):
            for x in range(mask.shape[1]):
                if cosmic_ray_mask[y, x]:
                    x0 = int(x)

                    y0 = y
                    y1 = y + 1 + settings.cosmic_ray_parallel_buffer

                    y1 = mask.shape[0] if y1 > mask.shape[0] else y1

                    mask[y0:y1, x] = True

                    x1 = int(x + 1 + settings.cosmic_ray_serial_buffer)

                    x1 = mask.shape[1] if x1 > mask.shape[1] else x1

                    mask[y, x0:x1] = True

                    y0 = y
                    y1 = y + 1 + settings.cosmic_ray_diagonal_buffer
                    x1 = int(x + 1 + settings.cosmic_ray_diagonal_buffer)

                    y1 = mask.shape[0] if y1 > mask.shape[0] else y1
                    x1 = mask.shape[1] if x1 > mask.shape[1] else x1

                    mask[y0:y1, x0:x1] = True

        return mask

    @classmethod
    def from_fits(
        cls, file_path, pixel_scales, hdu=0, origin=(0.0, 0.0), resized_mask_shape=None
    ):
        """
        Loads the image from a .fits file.

        Parameters
        ----------
        file_path
            The full path of the fits file.
        hdu
            The HDU number in the fits file containing the image image.
        pixel_scales or (float, float)
            The arc-second to pixel conversion factor of each pixel.
        """

        if type(pixel_scales) is not tuple:
            if type(pixel_scales) is float or int:
                pixel_scales = (float(pixel_scales), float(pixel_scales))

        mask = cls.manual(
            mask=aa.util.array_2d.numpy_array_2d_via_fits_from(
                file_path=file_path, hdu=hdu
            ),
            pixel_scales=pixel_scales,
            origin=origin,
        )

        if resized_mask_shape is not None:
            mask = mask.derive_mask.resized_from(new_shape=resized_mask_shape)

        return mask

    @classmethod
    def masked_fpr_and_eper_from(
        cls,
        mask: "Mask2D",
        layout: Layout2D,
        settings: "SettingsMask2D",
        pixel_scales: aa.type.PixelScales,
    ) -> "Mask2D":
        if settings.parallel_fpr_pixels is not None:
            parallel_fpr_mask = cls.masked_parallel_fpr_from(
                layout=layout, settings=settings, pixel_scales=pixel_scales
            )

            mask = mask + parallel_fpr_mask

        if settings.parallel_eper_pixels is not None:
            parallel_eper_mask = cls.masked_parallel_eper_from(
                layout=layout, settings=settings, pixel_scales=pixel_scales
            )

            mask = mask + parallel_eper_mask

        if settings.serial_fpr_pixels is not None:
            serial_fpr_mask = cls.masked_serial_fpr_from(
                layout=layout, settings=settings, pixel_scales=pixel_scales
            )

            mask = mask + serial_fpr_mask

        if settings.serial_eper_pixels is not None:
            serial_eper_mask = cls.masked_serial_eper_from(
                layout=layout, settings=settings, pixel_scales=pixel_scales
            )

            mask = mask + serial_eper_mask

        return mask

    @classmethod
    def masked_parallel_fpr_from(
        cls,
        layout: Layout2D,
        settings: "SettingsMask2D",
        pixel_scales: aa.type.PixelScales,
        invert: bool = False,
    ) -> "Mask2D":
        fpr_regions = layout.extract.parallel_fpr.region_list_from(
            settings=SettingsExtract(pixels=settings.parallel_fpr_pixels)
        )
        mask = np.full(layout.shape_2d, False)

        for region in fpr_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return Mask2D(mask=mask.astype("bool"), pixel_scales=pixel_scales)

    @classmethod
    def masked_parallel_eper_from(
        cls,
        layout: Layout2D,
        settings: "SettingsMask2D",
        pixel_scales: aa.type.PixelScales,
        invert: bool = False,
    ) -> "Mask2D":
        eper_regions = layout.extract.parallel_eper.region_list_from(
            settings=SettingsExtract(pixels=settings.parallel_eper_pixels)
        )

        mask = np.full(layout.shape_2d, False)

        for region in eper_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return Mask2D(mask=mask.astype("bool"), pixel_scales=pixel_scales)

    @classmethod
    def masked_serial_fpr_from(
        cls,
        layout: Layout2D,
        settings: "SettingsMask2D",
        pixel_scales: aa.type.PixelScales,
        invert: bool = False,
    ) -> "Mask2D":
        fpr_regions = layout.extract.serial_fpr.region_list_from(
            settings=SettingsExtract(pixels=settings.serial_fpr_pixels)
        )
        mask = np.full(layout.shape_2d, False)

        for region in fpr_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return Mask2D(mask=mask.astype("bool"), pixel_scales=pixel_scales)

    @classmethod
    def masked_serial_eper_from(
        cls,
        layout: Layout2D,
        settings: "SettingsMask2D",
        pixel_scales: aa.type.PixelScales,
        invert: bool = False,
    ) -> "Mask2D":
        eper_regions = layout.extract.serial_eper.region_list_from(
            settings=SettingsExtract(pixels=settings.serial_eper_pixels)
        )
        mask = np.full(layout.shape_2d, False)

        for region in eper_regions:
            mask[region.y0 : region.y1, region.x0 : region.x1] = True

        if invert:
            mask = np.invert(mask)

        return Mask2D(mask=mask.astype("bool"), pixel_scales=pixel_scales)



from autofit.non_linear.mock.mock_samples import MockSamples

from autoarray.fixtures import *

import autofit as af
import autocti as ac

import numpy as np

### Arctic ###


def make_trap_0():
    return ac.TrapInstantCapture(density=10.0, release_timescale=-1 / np.log(0.5))


def make_trap_1():
    return ac.TrapInstantCapture(density=8.0, release_timescale=-1 / np.log(0.2))


def make_traps_x1():
    return [make_trap_0()]


def make_traps_x2():
    return [make_trap_0(), make_trap_1()]


def make_ccd():
    return ac.CCDPhase(
        well_fill_power=0.5, full_well_depth=10000.0, well_notch_depth=1e-7
    )


def make_clocker_1d():
    return ac.Clocker1D(express=2)


def make_parallel_clocker_2d():
    return ac.Clocker2D(
        parallel_express=2, parallel_roe=ac.ROE(empty_traps_for_first_transfers=True)
    )


def make_serial_clocker_2d():
    return ac.Clocker2D(serial_express=2)


def make_parallel_serial_clocker_2d():
    return ac.Clocker2D(
        parallel_express=2,
        parallel_roe=ac.ROE(empty_traps_for_first_transfers=True),
        serial_express=2,
    )


### MASK ###


def make_mask_1d_7_unmasked():
    return ac.Mask1D.all_false(shape_slim=(7,), pixel_scales=(1.0,))


def make_mask_2d_7x7_unmasked():
    return ac.Mask2D.all_false(shape_native=(7, 7), pixel_scales=(1.0, 1.0))


### FRAMES ###


def make_image_7x7_native():
    return ac.Array2D.full(
        fill_value=1.0, shape_native=(7, 7), pixel_scales=(1.0, 1.0)
    ).native


def make_noise_map_7x7_native():
    return ac.Array2D.full(
        fill_value=2.0, shape_native=(7, 7), pixel_scales=(1.0, 1.0)
    ).native


### IMAGING ###


def make_imaging_7x7_frame():
    return ac.Imaging(
        data=make_image_7x7_native(),
        noise_map=make_noise_map_7x7_native(),
        name="mock_imaging_7x7_frame",
    )


### LINE DATASET ###


def make_layout_7():
    return ac.Layout1D(
        shape_1d=(7,), region_list=[(1, 5)], prescan=(0, 1), overscan=(6, 7)
    )


def make_data_7():
    return ac.Array1D.full(fill_value=1.0, shape_native=(7,), pixel_scales=1.0)


def make_noise_map_7():
    return ac.Array1D.full(fill_value=2.0, shape_native=(7,), pixel_scales=1.0)


def make_pre_cti_data_7():
    return ac.Array1D.full(fill_value=1.0, shape_native=(7,), pixel_scales=1.0)


def make_dataset_1d_7():
    return ac.Dataset1D(
        data=make_data_7(),
        noise_map=make_noise_map_7(),
        pre_cti_data=make_pre_cti_data_7(),
        layout=make_layout_7(),
    )


### CHARGE INJECTION FRAMES ###


def make_layout_ci_7x7():
    return ac.Layout2DCI(
        shape_2d=(7, 7),
        region_list=[(1, 5, 1, 5)],
        original_roe_corner=(1, 0),
        serial_overscan=(0, 6, 6, 7),
        serial_prescan=(0, 7, 0, 1),
        parallel_overscan=(6, 7, 1, 6),
    )


def make_ci_image_7x7():
    return ac.Array2D.full(fill_value=1.0, shape_native=(7, 7), pixel_scales=(1.0, 1.0))


def make_ci_noise_map_7x7():
    return ac.Array2D.full(fill_value=2.0, shape_native=(7, 7), pixel_scales=(1.0, 1.0))


def make_pre_cti_data_7x7():
    return ac.Array2D.full(
        shape_native=(7, 7), fill_value=10.0, pixel_scales=(1.0, 1.0)
    )


def make_ci_cosmic_ray_map_7x7():
    cosmic_ray_map = np.zeros(shape=(7, 7))

    return ac.Array2D.no_mask(values=cosmic_ray_map, pixel_scales=(1.0, 1.0))


def make_ci_noise_scaling_map_dict_7x7():
    return {
        "parallel_eper": ac.Array2D.ones(shape_native=(7, 7), pixel_scales=(1.0, 1.0)),
        "serial_eper": ac.Array2D.full(
            shape_native=(7, 7), fill_value=2.0, pixel_scales=(1.0, 1.0)
        ),
    }


### LINE DATASET FITS ###


def make_fit_1d_7():
    return ac.FitDataset1D(
        dataset=make_dataset_1d_7(), post_cti_data=make_dataset_1d_7().pre_cti_data + 1
    )


### CHARGE INJECTION IMAGING ###


def make_imaging_ci_7x7():
    return ac.ImagingCI(
        data=make_ci_image_7x7(),
        noise_map=make_ci_noise_map_7x7(),
        pre_cti_data=make_pre_cti_data_7x7(),
        cosmic_ray_map=make_ci_cosmic_ray_map_7x7(),
        noise_scaling_map_dict=make_ci_noise_scaling_map_dict_7x7(),
        layout=make_layout_ci_7x7(),
    )


### CHARGE INJECTION FITS ###


def make_hyper_noise_scalar_dict():
    return {
        "parallel_eper": ac.HyperCINoiseScalar(scale_factor=1.0),
        "serial_eper": ac.HyperCINoiseScalar(scale_factor=2.0),
    }


def make_fit_ci_7x7():
    return ac.FitImagingCI(
        dataset=make_imaging_ci_7x7(),
        post_cti_data=make_imaging_ci_7x7().pre_cti_data,
        hyper_noise_scalar_dict=make_hyper_noise_scalar_dict(),
    )


# ### PHASES ###


def make_samples_with_result():
    model = af.Collection(
        cti=af.Model(
            ac.CTI2D,
            parallel_trap_list=[ac.TrapInstantCapture],
            parallel_ccd=make_ccd(),
            serial_trap_list=[ac.TrapInstantCapture],
            serial_ccd=make_ccd(),
        ),
    )

    instance = model.instance_from_prior_medians()

    return MockSamples(
        model=model,
        max_log_likelihood_instance=instance,
        gaussian_tuples=[(1.0, 2.0)] * model.prior_count,
    )


def make_analysis_imaging_ci_7x7():
    return ac.AnalysisImagingCI(
        dataset=make_imaging_ci_7x7(), clocker=make_parallel_clocker_2d()
    )


### EUCLID DATA ####


def make_euclid_data():
    return np.zeros((2086, 2128))


### ACS DATA ####


def make_acs_ccd():
    return np.zeros((2068, 4144))


def make_acs_quadrant():
    return np.zeros((2068, 2072))



import copy

import autoarray as aa

from typing import Tuple


class SettingsImagingCI(aa.SettingsImaging):
    def __init__(
        self,
        parallel_pixels: Tuple[int, int] = None,
        serial_pixels: Tuple[int, int] = None,
    ):
        super().__init__()

        self.parallel_pixels = parallel_pixels
        self.serial_pixels = serial_pixels

    def modify_via_fit_type(self, is_parallel_fit, is_serial_fit):
        """
        Modify the settings based on the type of fit being performed where:

        - If the fit is a parallel only fit (is_parallel_fit=True, is_serial_fit=False) the serial_pixels are set to None
          and all other settings remain the same.

        - If the fit is a serial only fit (is_parallel_fit=False, is_serial_fit=True) the parallel_pixels are set to
          None and all other settings remain the same.

        - If the fit is a parallel and serial fit (is_parallel_fit=True, is_serial_fit=True) the *parallel_pixels* and
          *serial_pixels* are set to None and all other settings remain the same.

        These settings reflect the appropriate way to extract the charge injection imaging data for fits which use a
        parallel only CTI model, serial only CTI model or fit both.

        Parameters
        ----------
        is_parallel_fit
            If True, the CTI model that is used to fit the charge injection data includes a parallel CTI component.
        is_serial_fit
            If True, the CTI model that is used to fit the charge injection data includes a serial CTI component.
        """

        settings = copy.copy(self)

        if is_parallel_fit:
            settings.serial_pixels = None

        if is_serial_fit:
            settings.parallel_pixels = None

        return settings



import autoarray.plot as aplt


class GetVisuals1D(aplt.GetVisuals1D):
    def __init__(self, include: aplt.Include1D, visuals: aplt.Visuals1D):
        """
        Class which gets 1D attributes and adds them to a `Visuals1D` objects, such that they are plotted on 1D figures.

        For a visual to be extracted and added for plotting, it must have a `True` value in its corresponding entry in
        the `Include1D` object. If this entry is `False`, the `GetVisuals1D.get` method returns a None and the attribute
        is omitted from the plot.

        The `GetVisuals1D` class adds new visuals to a pre-existing `Visuals1D` object that is passed to its `__init__`
        method. This only adds a new entry if the visual are not already in this object.

        Parameters
        ----------
        include
            Sets which 1D visuals are included on the figure that is to be plotted (only entries which are `True`
            are extracted via the `GetVisuals1D` object).
        visuals
            The pre-existing visuals of the plotter which new visuals are added too via the `GetVisuals1D` class.
        """
        super().__init__(include=include, visuals=visuals)
